/*
Module : ShellListCtrl2.cpp
Purpose: Implementation for the CShellListCtrl2 class
Created: PJN / 25-07-2020
History: PJN / 25-07-2020 1. Initial creation
         PJN / 23-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable
                          declarations.
         PJN / 04-06-2023 1. Reworked CShellListCtrl2::OnGetItemText to use IShellItem2
                          2. Reworked CShellListCtrl2::OnCompareItems to use IShellItem2
                          3. Reworked the logic in CShellListCtrl2::OnCompareItems to behave
                          more similar to how Windows Explorer behaves.
         PJN / 11-12-2024 1. Added missing SAL to CShellListCtrl2 class
                          2. Added OLE Drag Drop support to CShellListCtrl2
         PJN / 12-12-2024 1. Added Right mouse OLE Drag Drop support to CShellListCtrl2 class
                          2. Removed logic to auto expand folders when dragging files into
                          CShellListCtrl2 to be consistent with modern Windows Explorer.
                          3. Updated CShellListCtrl2::OnBeginDrag to use SHDoDragDrop.
         PJN / 14-12-2024 1. Updated CShellListCtrl2 class to use new CShellCtrlScopedRedraw class
                          2. CShellListCtrl2::OnBeginDrag now does not do a Refresh after the 
                          operation completes.
                          3. Updated CShellListCtrl2 class to use std::filesystem::path class.

Copyright (c) 2020 - 2024 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "resource.h"
#include "ShellListCtrl2.h"
#include "ShellTreeCtrl2.h"
#include "ShellCtrl2ScopedRedraw.h"
#include <Shobjidl.h>
#include <propkey.h>

#ifndef _FILESYSTEM_
#pragma message("To avoid this message, put filesystem in your pre compiled header (normally stdafx.h)")
#include <filesystem>
#endif //#ifndef _FILESYSTEM_


//////////////////// Defines / Locals /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

#pragma warning(suppress: 26426)
const UINT WM_SHELLLISTCTRL2_AFTER_SHELL_COMMAND{::RegisterWindowMessage(CSHELLLISTCTRL2_WM_SHELLLISTCTRL2_AFTER_SHELL_COMMAND)};
#pragma warning(suppress: 26426)
const UINT WM_SHELLLISTCTRL2_NOTIFICATION_CHANGE{::RegisterWindowMessage(CSHELLLISTCTRL2_WM_SHELLLISTCTRL2_NOTIFICATION_CHANGE)};
#pragma warning(suppress: 26426)
const UINT WM_SHELLLISTCTRL2_AFTER_OLEDROP{::RegisterWindowMessage(CSHELLLISTCTRL2_WM_SHELLLISTCTRL2_AFTER_OLEDROP)};



//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26409 26440 26433 26477)
IMPLEMENT_DYNCREATE(CShellListCtrl2, SHELLLISTCTRL2_BASE_CLASS)

#pragma warning(suppress: 26455)
CShellListCtrl2::CShellListCtrl2() : m_EnumObjectFlags{SHCONTF_FOLDERS | SHCONTF_NONFOLDERS | SHCONTF_INCLUDEHIDDEN | SHCONTF_INCLUDESUPERHIDDEN},
                                     m_hRelatedTree{nullptr},
                                     m_bShellContextMenu{true},
                                     m_bNoSelectedItemContextMenu{true},
                                     m_bIsDesktop{false},
                                     m_bNoNotify{false},
                                     m_bAllowDragDrop{false},
                                     m_bAllowDoShellDefault{false},
                                     m_dwDriveHideFlags{0},
                                     m_dwFileHideFlags{FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_TEMPORARY},
                                     m_dwFolderHideFlags{FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_TEMPORARY},
                                     m_rgbCompressed{RGB(0, 0, 255)},
                                     m_bShowCompressedUsingDifferentColor{true},
                                     m_rgbEncrypted{RGB(98, 146, 13)},
                                     m_bShowEncryptedUsingDifferentColor{true},
                                     m_bShowSharedUsingDifferentIcon{true},
                                     m_bShowLinkedItemsUsingDifferentIcon{true},
                                     m_bRefreshAfterShellCommand{true},
                                     m_pMonitoringThread{nullptr},
                                     m_bAutoRefresh{false},
                                     m_bShowFiles{true},
                                     m_pOLEDropTarget{nullptr},
                                     m_bCanAcceptData{false},
                                     m_bLeftMouseDrag{true},
                                     m_nItemDrop{-1},
                                     m_nLastMovementTime{0},
                                     m_nLastScrollTime{0},
                                     m_lastMouseLoc{INT_MIN, INT_MIN},
                                     m_nMouseHoverTime{400},
                                     m_bDropContainedDirectories{false},
                                     m_bDropContainedFiles{false},
                                     m_nDropTarget{-1},
                                     m_bDropTargetWasSelected{false}
{
  m_FilenameMasks.emplace_back(_T("*.*"));
}

CShellListCtrl2::~CShellListCtrl2()
{
  if (m_pOLEDropTarget != nullptr)
  {
#pragma warning(suppress: 26447)
    m_pOLEDropTarget->InternalRelease();
    m_pOLEDropTarget = nullptr;
  }
}


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CShellListCtrl2, SHELLLISTCTRL2_BASE_CLASS) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_CREATE()
  ON_WM_CONTEXTMENU()
  ON_WM_DESTROY()
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(LVN_DELETEITEM, &CShellListCtrl2::OnDeleteItem)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(NM_DBLCLK, &CShellListCtrl2::OnDblClk)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(NM_RETURN, &CShellListCtrl2::OnReturn)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(NM_CUSTOMDRAW, &CShellListCtrl2::OnCustomDraw)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(LVN_BEGINDRAG, &CShellListCtrl2::OnBeginDrag)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(LVN_BEGINRDRAG, &CShellListCtrl2::OnBeginRDrag)
  ON_REGISTERED_MESSAGE(WM_SHELLLISTCTRL2_NOTIFICATION_CHANGE, &CShellListCtrl2::OnNotificationChange)
END_MESSAGE_MAP()

#pragma warning(suppress: 26434)
int CShellListCtrl2::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
  if (__super::OnCreate(lpCreateStruct) == -1) //Let the base class do its thing
    return -1;
  if (!InitList())
    return -1;
  return 0;
}

HIMAGELIST CShellListCtrl2::GetShellImageList(_In_ bool bLarge)
{
  CString sWinDir;
  const UINT nGWD{GetWindowsDirectory(sWinDir.GetBuffer(_MAX_PATH), _MAX_PATH)};
  sWinDir.ReleaseBuffer();
  if (nGWD == 0)
#pragma warning(suppress: 26487)
    return nullptr;
  SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
  auto hImageList{reinterpret_cast<HIMAGELIST>(SHGetFileInfo(sWinDir, 0, &sfi, sizeof(SHFILEINFO), SHGFI_SYSICONINDEX | (bLarge ? 0 : SHGFI_SMALLICON)))};
#pragma warning(suppress: 26487)
  return hImageList;
}

HRESULT CShellListCtrl2::LockCurrentFolder(_Inout_ ATL::CComPtr<IShellFolder>& pParentFolder, _In_ const CPIDL& pidlRel, _In_ const CPIDL& pidlFQ)
{
  HRESULT hr{E_FAIL};
  if (m_psfCurFolder != nullptr)
    m_psfCurFolder.Release();
  m_pidlCurFQ.Free();

  if ((pParentFolder != nullptr) && pidlRel.GetItemCount())
  {
    ASSERT(pidlRel.operator bool());
#pragma warning(suppress: 26490)
    hr = pParentFolder->BindToObject(pidlRel, nullptr, IID_IShellFolder, reinterpret_cast<LPVOID*>(&m_psfCurFolder));
    m_bIsDesktop = false;
  }
  else
  {
    hr = SHGetDesktopFolder(&m_psfCurFolder);
    m_bIsDesktop = true;
  }
  if (SUCCEEDED(hr))
    m_pidlCurFQ = pidlFQ;
  return hr;
}

HRESULT CShellListCtrl2::DisplayFolder(_In_opt_ CShellItemInfo2* pItemInfo)
{
  HRESULT hr{E_FAIL};
  if (pItemInfo != nullptr)
  {
    hr = LockCurrentFolder(pItemInfo->m_pParentFolder, pItemInfo->m_pidlRel, pItemInfo->m_pidlFQ);
    if (FAILED(hr))
      return hr;
  }
  KillMonitoringThread();
  DeleteAllItems();
  if (m_psfCurFolder != nullptr)
  {
    CWaitCursor wait;
    CShellCtrlScopedRedraw redraw(this);
    hr = EnumObjects(m_psfCurFolder, m_pidlCurFQ);
    const CMFCHeaderCtrl& headerCtrl{GetHeaderCtrl()};
    if (headerCtrl.GetSafeHwnd() != nullptr)
    {
      const int nSortColumn{headerCtrl.GetSortColumn()};
      if (nSortColumn != -1)
      {
        BOOL bSortAscending{TRUE};
        if (nSortColumn != -1)
          bSortAscending = headerCtrl.IsAscending();
        Sort(nSortColumn, bSortAscending);
      }
    }
  }

  //Setup notifications if required
  CString sPath;
  if (SUCCEEDED(hr) && m_bAutoRefresh && m_pidlCurFQ.ToPath(sPath) && CanHandleChangeNotifications(sPath))
#pragma warning(suppress: 26489)
    CreateMonitoringThread(sPath);

  //Inform the related tree control if necessary
  if (SUCCEEDED(hr) && (pItemInfo != nullptr))
  {
    CShellTreeCtrl2* pTree{GetRelatedTree()};
    if ((pTree != nullptr) && !m_bNoNotify)
      pTree->SelectPath(m_pidlCurFQ);
    const CWnd* pParent{GetParent()};
    if (pParent != nullptr)
      pParent->SendMessage(WM_SHELLLISTCTRL2_AFTER_SHELL_COMMAND);
  }

  return hr;
}

HRESULT CShellListCtrl2::DisplayParentFolder()
{
  if (!m_pidlCurFQ.operator bool())
    return E_FAIL;
  CShellItemInfo2 info;
  HRESULT hr{m_pidlCurFQ.GetParent(info.m_pidlFQ)};
  if (FAILED(hr))
    return hr;
  if (info.m_pidlFQ.GetItemCount() == 0) //Desktop
    hr = DisplayFolder(&info);
  else
  {
    hr = SHGetDesktopFolder(&info.m_pParentFolder);
    if (FAILED(hr))
      return hr;
    info.m_pidlRel = info.m_pidlFQ;
    hr = DisplayFolder(&info);
  }
  return hr;
}

HRESULT CShellListCtrl2::DisplayFolder(_In_ const CPIDL& pidl)
{
  CShellItemInfo2 info;
  info.m_pidlRel = pidl;
  const HRESULT hr{SHGetDesktopFolder(&info.m_pParentFolder)};
  if (FAILED(hr))
    return hr;
  info.m_pidlFQ = info.m_pidlRel;
  return DisplayFolder(&info);
}

HRESULT CShellListCtrl2::DisplayFolder(_In_z_ LPCTSTR lpszPath)
{
  //Validate our parameters
  ASSERT(lpszPath != nullptr);

  CPIDL pidl;
  const HRESULT hr{pidl.FromPath(lpszPath)};
  if (FAILED(hr))
    return hr;
  return DisplayFolder(pidl);
}

HRESULT CShellListCtrl2::Refresh()
{
  return DisplayFolder(static_cast<CShellItemInfo2*>(nullptr));
}

int CShellListCtrl2::AddItem(_In_ ATL::CComPtr<IShellFolder>& pParentFolder, _In_ const CPIDL& pidl, _In_ SFGAOF attribs, _In_ const CPIDL& pidlParent)
{
  //fill in the LVITEM structure for this item
  LVITEM lvItem{};
  lvItem.iItem = GetItemCount();
  lvItem.mask = LVIF_PARAM | LVIF_TEXT | LVIF_IMAGE | LVIF_STATE;

  //Put the private information in the lParam
#pragma warning(suppress: 26400 26403 26409)
  auto pItem{new CShellItemInfo2};
  pItem->m_pidlRel = pidl;
  pItem->m_pidlFQ = pidlParent + pidl;
  pItem->m_Attribs = attribs;
  pItem->m_pParentFolder = pParentFolder;
#pragma warning(suppress: 26490)
  lvItem.lParam = reinterpret_cast<LPARAM>(pItem);
  lvItem.iImage = OnGetItemIcon(GetItemCount(), pItem);

  if (m_bShowSharedUsingDifferentIcon && (attribs & SFGAO_SHARE))
  {
    lvItem.mask |= LVIF_STATE;
    lvItem.stateMask |= LVIS_OVERLAYMASK;
    lvItem.state |= INDEXTOOVERLAYMASK(1); //1 is the index for the shared overlay image
  }
  if (m_bShowLinkedItemsUsingDifferentIcon && (attribs & SFGAO_LINK))
  {
    lvItem.mask |= LVIF_STATE;
    lvItem.stateMask |= LVIS_OVERLAYMASK;
    lvItem.state |= INDEXTOOVERLAYMASK(2); //2 is the index for the linked / shortcuts overlay image
  }
  if (attribs & SFGAO_GHOSTED)
  {
    lvItem.mask |= LVIF_STATE;
    lvItem.stateMask |= LVIS_CUT;
    lvItem.state |= LVIS_CUT;
  }

  //Finally add the item
  const int nItem{InsertItem(&lvItem)};
  if (nItem == -1)
    return nItem;

  //Set columns
  for (int nColumn{static_cast<int>(Columns::NameColumn)}; nColumn<=static_cast<int>(Columns::SizeColumn); nColumn++)
    SetItemText(nItem, nColumn, OnGetItemText(nItem, nColumn, pItem));

  //Call the virtual OnItemAdded method to allow derived classes to customize our behaviour
  OnItemAdded(nItem);

  return nItem;
}

#pragma warning(suppress: 26440)
SFGAOF CShellListCtrl2::GetAttributesOfFlags()
{
  //Form the flags we will pass to IShellFolder::GetAttributesOf
  SFGAOF attribs{SFGAO_GHOSTED};
  if (m_bShowSharedUsingDifferentIcon)
    attribs |= SFGAO_SHARE;
  if (m_bShowCompressedUsingDifferentColor)
    attribs |= SFGAO_COMPRESSED;
  if (m_bShowEncryptedUsingDifferentColor)
    attribs |= SFGAO_ENCRYPTED;
  if (m_bShowLinkedItemsUsingDifferentIcon)
    attribs |= SFGAO_LINK;
  return attribs;
}

HRESULT CShellListCtrl2::EnumObjects(_In_ ATL::CComPtr<IShellFolder>& pParentFolder, _In_ const CPIDL& pidlParent)
{
  //Use IShellFolder and IEnumIDList to enumerate the items in the parent folder
  ATL::CComPtr<IEnumIDList> pEnum;
  const HRESULT hr{pParentFolder->EnumObjects(nullptr, m_EnumObjectFlags, &pEnum)};
  if ((FAILED(hr)) || (pEnum == nullptr))
    return hr;
  CPIDL pidl;
  DWORD dwFetched{1};
  while (SUCCEEDED(pEnum->Next(1, pidl, &dwFetched)) && dwFetched)
  {
    //Add the item if necessary
    SFGAOF attribs{GetAttributesOfFlags()};
    if (SUCCEEDED(pParentFolder->GetAttributesOf(1, pidl, &attribs)) && CanDisplay(pidlParent, pidl))
      AddItem(pParentFolder, pidl, attribs, pidlParent);
    //Prepare for the next time around
    dwFetched = 0;
    pidl.Free();
  }
  return hr;
}

_Success_(return != false) bool CShellListCtrl2::DoShellDefault(_In_ int nItem, _In_ bool bOpenInControlIfPossible)
{
  LVITEM lvItem{};
  lvItem.mask = LVIF_PARAM;
  lvItem.iItem = nItem;
  if (!GetItem(&lvItem))
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26490)
  auto pItem{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
#pragma warning(suppress: 26489)
  if ((pItem == nullptr) || (!pItem->m_pidlRel.operator bool()))
  {
    ASSERT(FALSE);
    return false;
  }

  //If specified element is a folder, try to display it
  ASSERT(pItem->m_pParentFolder != nullptr);
  ULONG ulAttrs{SFGAO_FOLDER};
#pragma warning(suppress: 26489)
  pItem->m_pParentFolder->GetAttributesOf(1, pItem->m_pidlRel, &ulAttrs);
  if ((ulAttrs & SFGAO_FOLDER) && bOpenInControlIfPossible)
  {
#pragma warning(suppress: 26486)
    return SUCCEEDED(DisplayFolder(pItem));
  }

  if (!m_bAllowDoShellDefault)
    return false;

  //Invoke the default menu command
  ATL::CComPtr<IContextMenu> pcm;
#pragma warning(suppress: 26490)
  HRESULT hr{pItem->m_pParentFolder->GetUIObjectOf(GetSafeHwnd(), 1, pItem->m_pidlRel, IID_IContextMenu, nullptr, reinterpret_cast<void**>(&pcm))};
  if (FAILED(hr))
    return false;
  CMenu popup;
  if (!popup.CreatePopupMenu())
    return false;
  hr = pcm->QueryContextMenu(popup, 0, 1, 0x7FFFF, CMF_DEFAULTONLY | CMF_EXPLORE);
  if (FAILED(hr))
    return false;
  const UINT nIDCmd{popup.GetDefaultItem(0, FALSE)};
#pragma warning(suppress: 26472)
  if ((nIDCmd == 0) || (nIDCmd == static_cast<UINT>(-1)))
    return false;
  const CWnd* pParent{GetParent()};
  if (pParent == nullptr)
  {
    ASSERT(FALSE);
    return false;
  }
  CMINVOKECOMMANDINFO cmi{};
  cmi.cbSize = sizeof(CMINVOKECOMMANDINFO);
  cmi.hwnd = pParent->GetSafeHwnd();
#pragma warning(suppress: 26472 26490)
  cmi.lpVerb = reinterpret_cast<LPCSTR>(static_cast<INT_PTR>(nIDCmd - 1));
  cmi.nShow = SW_SHOWNORMAL;
  hr = pcm->InvokeCommand(&cmi);
  if (SUCCEEDED(hr))
#pragma warning(suppress: 26472)
    pParent->SendMessage(WM_SHELLLISTCTRL2_AFTER_SHELL_COMMAND, static_cast<WPARAM>(nIDCmd));
  return SUCCEEDED(hr);
}

#pragma warning(suppress: 26429 26434 26440)
void CShellListCtrl2::OnDeleteItem(NMHDR* pNMHDR, LRESULT* pResult)
{
#pragma warning(suppress: 26490)
  auto pNMListView{reinterpret_cast<NM_LISTVIEW*>(pNMHDR)};
#pragma warning(suppress: 26496)
  ASSERT(pNMListView != nullptr);
  ASSERT(pResult != nullptr);
#pragma warning(suppress: 26429 26490)
  auto pItem{reinterpret_cast<CShellItemInfo2*>(pNMListView->lParam)};
  ASSERT(pItem != nullptr);
#pragma warning(suppress: 26401 26409 26486 26489)
  delete pItem;
  *pResult = 0;
}

#pragma warning(suppress: 26429)
void CShellListCtrl2::OnDblClk(NMHDR* /*pNMHDR*/, LRESULT* pResult)
{
  //Validate our parameters
  ASSERT(pResult != nullptr);

  const int nItem{GetNextItem(-1, LVNI_FOCUSED)};
  if (nItem != -1)
    DoShellDefault(nItem);
  *pResult = 0;
}

#pragma warning(suppress: 26429)
void CShellListCtrl2::OnReturn(NMHDR* /*pNMHDR*/, LRESULT* pResult)
{
  //Validate our parameters
  ASSERT(pResult != nullptr);

  const int nItem{GetNextItem(-1, LVNI_FOCUSED)};
  if (nItem != -1)
    DoShellDefault(nItem);
  *pResult = 0;
}

_Success_(return != false) bool CShellListCtrl2::GetItemPath(_In_ int nItem, _Inout_ CString& sPath) const
{
  sPath.Empty(); //Default the output parameter

  LVITEM lvItem{};
  lvItem.mask = LVIF_PARAM;
  lvItem.iItem = nItem;
  if (!GetItem(&lvItem))
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26490)
  const CShellItemInfo2* pItem{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
#pragma warning(suppress: 26489)
  if ((pItem == nullptr) || (!pItem->m_pidlFQ.operator bool()))
    return false;
#pragma warning(suppress: 26489)
  return pItem->m_pidlFQ.ToPath(sPath);
}

_Success_(return != false) bool CShellListCtrl2::GetItemPIDL(_In_ int nItem, _Inout_ CPIDL& pidl) const
{
  LVITEM lvItem{};
  lvItem.mask = LVIF_PARAM;
  lvItem.iItem = nItem;
  if (!GetItem(&lvItem))
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26434 26490)
  const CShellItemInfo2* pItem{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
  if (pItem == nullptr)
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26489)
  pidl = pItem->m_pidlFQ;
  return true;
}

_Success_(return != false) bool CShellListCtrl2::GetCurrentFolder(_Inout_ CString& sPath) const
{
  sPath.Empty(); //Default the output parameter
  if (!m_pidlCurFQ.operator bool())
    return false;
  return m_pidlCurFQ.ToPath(sPath);
}

_Success_(return != false) bool CShellListCtrl2::GetCurrentFolderName(_Inout_ CString& sName) const
{
  sName.Empty(); //Default the output parameter
  if (!m_pidlCurFQ.operator bool())
    return false;
  SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
  if (!SHGetFileInfo(reinterpret_cast<LPCTSTR>(m_pidlCurFQ.m_pPIDL), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_DISPLAYNAME))
    return false;
#pragma warning(suppress: 26485)
  sName = sfi.szDisplayName;
  return true;
}

CString CShellListCtrl2::OnGetItemText(_In_ int /*nItem*/, _In_ int nColumn, _In_ CShellItemInfo2* pItem)
{
  //Validate our parameters
  ASSERT(pItem != nullptr);

  switch (nColumn)
  {
    case static_cast<int>(Columns::NameColumn):
    {
      SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
      if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(pItem->m_pidlFQ.m_pPIDL), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_DISPLAYNAME))
#pragma warning(suppress: 26485)
        return sfi.szDisplayName;
      break;
    }
    case static_cast<int>(Columns::TypeColumn):
    {
      SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
      if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(pItem->m_pidlFQ.m_pPIDL), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_TYPENAME))
#pragma warning(suppress: 26485)
        return sfi.szTypeName;
      break;
    }
    case static_cast<int>(Columns::SizeColumn):
    {
      ATL::CComPtr<IShellItem2> pShellItem;
#pragma warning(suppress: 26490)
      HRESULT hr{SHCreateItemFromIDList(pItem->m_pidlFQ, IID_IShellItem2, reinterpret_cast<void**>(&pShellItem))};
      if (FAILED(hr))
        break;
      if (!pShellItem)
        break;
      ULONGLONG nSize{0};
      hr = pShellItem->GetUInt64(PKEY_Size, &nSize);
      if (FAILED(hr))
        break;
      CString sText;
      OnFormatFileSize(nSize, sText);
      return sText;
      break;
    }
    case static_cast<int>(Columns::DateModifiedColumn):
    {
      ATL::CComPtr<IShellItem2> pShellItem;
#pragma warning(suppress: 26490)
      HRESULT hr{SHCreateItemFromIDList(pItem->m_pidlFQ, IID_IShellItem2, reinterpret_cast<void**>(&pShellItem))};
      if (FAILED(hr))
        break;
      if (!pShellItem)
        break;
      FILETIME ftModified{};
      hr = pShellItem->GetFileTime(PKEY_DateModified, &ftModified);
      if (FAILED(hr))
        break;
      CString sText;
      OnFormatFileDate(ftModified, sText);
      return sText;
      break;
    }
    default:
    {
      ASSERT(FALSE);
      break;
    }
  }
  return _T("");
}

#pragma warning(suppress: 26440)
int CShellListCtrl2::OnGetItemIcon(_In_ int /*nItem*/, _In_ CShellItemInfo2* pItem)
{
  if (pItem == nullptr)
  {
    ASSERT(FALSE);
    return -1;
  }
  SHFILEINFO sfi{};
  int iIcon{-1};
#pragma warning(suppress: 26490)
  if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(pItem->m_pidlFQ.m_pPIDL), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_SYSICONINDEX | SHGFI_SMALLICON | SHGFI_LINKOVERLAY))
    iIcon = sfi.iIcon;
  return iIcon;
}

int CShellListCtrl2::OnCompareItems(_In_ LPARAM lParam1, _In_ LPARAM lParam2, _In_ int nColumn)
{
#pragma warning(suppress: 26490)
  auto pItem1{reinterpret_cast<CShellItemInfo2*>(lParam1)};
  ASSERT(pItem1 != nullptr);
#pragma warning(suppress: 26490)
  auto pItem2{reinterpret_cast<CShellItemInfo2*>(lParam2)};
  ASSERT(pItem2 != nullptr);

  int nResult{0};
  switch (nColumn)
  {
    case static_cast<int>(Columns::NameColumn):
    {
      //Validate our parameters
      ASSERT(pItem1->m_pParentFolder != nullptr);

#pragma warning(suppress: 26489)
      const HRESULT hr{pItem1->m_pParentFolder->CompareIDs(0, pItem1->m_pidlRel, pItem2->m_pidlRel)};
      if (FAILED(hr))
        return 0;
#pragma warning(suppress: 26472)
      nResult = static_cast<short>(SCODE_CODE(GetScode(hr)));
      break;
    }
    case static_cast<int>(Columns::TypeColumn):
    {
      //Validate our parameters
      ASSERT(pItem1->m_pParentFolder != nullptr);
      ASSERT(pItem2->m_pParentFolder != nullptr);

      //Determine if the two items are folders
      ULONG ulAttrs{SFGAO_FOLDER};
#pragma warning(suppress: 26489)
      pItem1->m_pParentFolder->GetAttributesOf(1, pItem1->m_pidlRel, &ulAttrs);
      const bool bFolder1 = (ulAttrs & SFGAO_FOLDER);
      ulAttrs = SFGAO_FOLDER;
      pItem2->m_pParentFolder->GetAttributesOf(1, pItem2->m_pidlRel, &ulAttrs);
      const bool bFolder2 = (ulAttrs & SFGAO_FOLDER);

      if ((bFolder1 && bFolder2) || (!bFolder1 && !bFolder2))
      {
        SHFILEINFO sfi1{};
        SHFILEINFO sfi2{};
#pragma warning(suppress: 26489 26490)
        if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(pItem1->m_pidlFQ.m_pPIDL), 0, &sfi1, sizeof(sfi1), SHGFI_PIDL | SHGFI_TYPENAME) &&
            SHGetFileInfo(reinterpret_cast<LPCTSTR>(pItem2->m_pidlFQ.m_pPIDL), 0, &sfi2, sizeof(sfi2), SHGFI_PIDL | SHGFI_TYPENAME))
#pragma warning(suppress: 26485)
          nResult = lstrcmpi(sfi1.szTypeName, sfi2.szTypeName);

        //Windows Explorer seems to compare folders by name if there is a match when comparing by type
        if (nResult == 0)
          nResult = OnCompareItems(lParam1, lParam2, static_cast<int>(Columns::NameColumn));
      }
      else if (bFolder1)
      {
        //Emulate the behaviour of Windows Explorer
        if (!m_bAscending)
          nResult = 1;
        else
          nResult = -1;
      }
      else
      {
        //Emulate the behaviour of Windows Explorer
        if (!m_bAscending)
          nResult = -1;
        else
          nResult = 1;
      }
      break;
    }
    case static_cast<int>(Columns::SizeColumn):
    {
      //Validate our parameters
      ASSERT(pItem1->m_pParentFolder != nullptr);
      ASSERT(pItem2->m_pParentFolder != nullptr);

      //Determine if the two items are folders
      ULONG ulAttrs{SFGAO_FOLDER};
#pragma warning(suppress: 26489)
      pItem1->m_pParentFolder->GetAttributesOf(1, pItem1->m_pidlRel, &ulAttrs);
      const bool bFolder1 = (ulAttrs & SFGAO_FOLDER);
      ulAttrs = SFGAO_FOLDER;
      pItem2->m_pParentFolder->GetAttributesOf(1, pItem2->m_pidlRel, &ulAttrs);
      const bool bFolder2 = (ulAttrs & SFGAO_FOLDER);

      if (bFolder1 && bFolder2)
      {
        //Windows Explorer seems to compare folders ascending by name when sorting by size
        nResult = OnCompareItems(lParam1, lParam2, static_cast<int>(Columns::NameColumn));
        if (!m_bAscending)
          nResult = -nResult;
      }
      else if (!bFolder1 && !bFolder2)
      {
        ATL::CComPtr<IShellItem2> pShellItem1;
#pragma warning(suppress: 26490)
        HRESULT hr{SHCreateItemFromIDList(pItem1->m_pidlFQ, IID_IShellItem2, reinterpret_cast<void**>(&pShellItem1))};
        if (FAILED(hr))
        {
          nResult = 1;
          break;
        }
        ULONGLONG nSize1{0};
        hr = pShellItem1->GetUInt64(PKEY_Size, &nSize1);
        if (FAILED(hr))
        {
          nResult = 1;
          break;
        }
        ATL::CComPtr<IShellItem2> pShellItem2;
#pragma warning(suppress: 26490)
        hr = SHCreateItemFromIDList(pItem2->m_pidlFQ, IID_IShellItem2, reinterpret_cast<void**>(&pShellItem2));
        if (FAILED(hr))
        {
          nResult = -1;
          break;
        }
        ULONGLONG nSize2{0};
        hr = pShellItem2->GetUInt64(PKEY_Size, &nSize2);
        if (FAILED(hr))
        {
          nResult = -1;
          break;
        }
        nResult = (nSize1 < nSize2) ? -1 : (nSize1 > nSize2) ? 1 : 0;
      }
      else if (bFolder1)
        nResult = -1;
      else
        nResult = 1;
      break;
    }
    case static_cast<int>(Columns::DateModifiedColumn):
    {
      //Validate our parameters
      ASSERT(pItem1->m_pParentFolder != nullptr);
      ASSERT(pItem2->m_pParentFolder != nullptr);

      //Determine if the two items are folders
      ULONG ulAttrs{SFGAO_FOLDER};
#pragma warning(suppress: 26489)
      pItem1->m_pParentFolder->GetAttributesOf(1, pItem1->m_pidlRel, &ulAttrs);
      const bool bFolder1 = (ulAttrs & SFGAO_FOLDER);
      ulAttrs = SFGAO_FOLDER;
      pItem2->m_pParentFolder->GetAttributesOf(1, pItem2->m_pidlRel, &ulAttrs);
      const bool bFolder2 = (ulAttrs & SFGAO_FOLDER);

      if ((bFolder1 && bFolder2) || (!bFolder1 && !bFolder2))
      {
        ATL::CComPtr<IShellItem2> pShellItem1;
#pragma warning(suppress: 26490)
        HRESULT hr{SHCreateItemFromIDList(pItem1->m_pidlFQ, IID_IShellItem2, reinterpret_cast<void**>(&pShellItem1))};
        if (FAILED(hr))
        {
          nResult = 1;
          break;
        }
        FILETIME ft1{};
        hr = pShellItem1->GetFileTime(PKEY_DateModified, &ft1);
        if (FAILED(hr))
        {
          nResult = 1;
          break;
        }
        ATL::CComPtr<IShellItem2> pShellItem2;
#pragma warning(suppress: 26490)
        hr = SHCreateItemFromIDList(pItem2->m_pidlFQ, IID_IShellItem2, reinterpret_cast<void**>(&pShellItem2));
        if (FAILED(hr))
        {
          nResult = -1;
          break;
        }
        FILETIME ft2{};
        hr = pShellItem2->GetFileTime(PKEY_DateModified, &ft2);
        if (FAILED(hr))
        {
          nResult = -1;
          break;
        }
        nResult = CompareFileTime(&ft1, &ft2);
      }
      else if (bFolder1)
        nResult = -1;
      else
        nResult = 1;
      break;
    }
    default:
    {
      ASSERT(FALSE);
      break;
    }
  }
  return nResult;
}

_Success_(return != false) bool CShellListCtrl2::OnSetColumns()
{
  //Delete all columns
#pragma warning(suppress: 26429)
  const CMFCHeaderCtrl& headerCtrl{GetHeaderCtrl()};
  if (headerCtrl.GetSafeHwnd() != nullptr)
  {
    const int nColumnCount{headerCtrl.GetItemCount()};
    for (int i{0}; i<nColumnCount; i++)
      DeleteColumn(0);
  }

  //Add the new columns
  constexpr std::array<int, 4> arColumnNameIDs
  { {
    IDS_SHELLCTRLS2_NAME,
    IDS_SHELLCTRLS2_DATE_MODIFIED,
    IDS_SHELLCTRLS2_TYPE,
    IDS_SHELLCTRLS2_SIZE
  } };
  for (int nColumn{static_cast<int>(Columns::NameColumn)}; nColumn<=static_cast<int>(Columns::SizeColumn); nColumn++)
  {
    CString sColumnName;
#pragma warning(suppress: 26446 26482)
    if (!sColumnName.LoadString(arColumnNameIDs[nColumn]))
      return false;
    const int nFormat{(nColumn == static_cast<int>(Columns::SizeColumn)) ? LVCFMT_RIGHT : LVCFMT_LEFT};
    int nWidth{0};
    if (nColumn == static_cast<int>(Columns::NameColumn))
      nWidth = 300;
    else if (nColumn == static_cast<int>(Columns::SizeColumn))
      nWidth = 100;
    else
      nWidth = 200;
    InsertColumn(nColumn, sColumnName, nFormat, nWidth, nColumn);
  }
  return true;
}

void CShellListCtrl2::PreSubclassWindow()
{
  //Let the base class do its thing
  __super::PreSubclassWindow();

  const _AFX_THREAD_STATE* pThreadState{AfxGetThreadState()};
  if (pThreadState->m_pWndInit == nullptr)
    InitList();
}

#pragma warning(suppress: 26434)
_Success_(return != false) bool CShellListCtrl2::InitList()
{
  ModifyStyle(0, LVS_SHAREIMAGELISTS);

  //Set shell images
  SetImageList(CImageList::FromHandle(GetShellImageList(true)), LVSIL_NORMAL);
  SetImageList(CImageList::FromHandle(GetShellImageList(false)), LVSIL_SMALL);

  //Add columns
  OnSetColumns();

  //Create the various drag drop objects
  if (!CreateDragDropComponents())
    return false;

  if (m_psfCurFolder == nullptr)
  {
    //Display desktop
    CShellItemInfo2 info;
    if (FAILED(SHGetKnownFolderIDList(FOLDERID_Desktop, 0, nullptr, info.m_pidlFQ)))
      return false;
    DisplayFolder(&info);
  }

  return true;
}

_Success_(return != false) bool CShellListCtrl2::CreateDragDropComponents()
{
  if (m_pOLEDropTarget == nullptr)
  {
#pragma warning(suppress: 26409)
    m_pOLEDropTarget = new CShellCtrl2DropTarget(this);
    if (m_bAllowDragDrop)
    {
      if (!m_pOLEDropTarget->Register(this))
        return false;
    }
  }
  return true;
}

void CShellListCtrl2::SetAllowDragDrop(_In_ bool bAllowDragDrop)
{
  if (m_bAllowDragDrop && !bAllowDragDrop)
  {
    //Validate our parameters
    ASSERT(m_pOLEDropTarget != nullptr);

    m_bAllowDragDrop = bAllowDragDrop;
    m_pOLEDropTarget->Revoke();
  }
  else if (!m_bAllowDragDrop && bAllowDragDrop)
  {
    //Validate our parameters
    ASSERT(m_pOLEDropTarget != nullptr);

    m_bAllowDragDrop = bAllowDragDrop;
    m_pOLEDropTarget->Register(this);
  }
}

void CShellListCtrl2::OnFormatFileSize(_In_ ULONGLONG nFileSize, _Inout_ CString& sSize)
{
  //Convert number to the system format:
  StrFormatKBSize(nFileSize, sSize.GetBuffer(255), 255);
  sSize.ReleaseBuffer();
}

void CShellListCtrl2::OnFormatFileDate(_In_ const FILETIME& ft, _Inout_ CString& sDate)
{
  const COleDateTime dateFile{ft};
  sDate = dateFile.Format();
}

#pragma warning(suppress: 26434)
void CShellListCtrl2::OnContextMenu(CWnd* /*pWnd*/, CPoint point)
{
  const UINT nSelItems{GetSelectedCount()};
  if ((nSelItems == 0) && m_bNoSelectedItemContextMenu)
    OnShowNothingSelectedContextMenu();
  else if (m_bShellContextMenu)
    OnShowShellContextMenu(point);
  else
    Default();
}

void CShellListCtrl2::OnShowNothingSelectedContextMenu()
{
  const UINT nSelItems{GetSelectedCount()};
  if (!m_bNoSelectedItemContextMenu)
    return;
  if (nSelItems)
    return;
  CMenu menu;
  if (!menu.LoadMenu(IDR_SHELLCTRLS2_POPUP_NOTHING_SELECTED))
  {
    ASSERT(FALSE);
    return;
  }
  CMenu* pPopup{menu.GetSubMenu(0)};
  if (pPopup == nullptr)
  {
    ASSERT(FALSE);
    return;
  }
  const DWORD dwStyle{GetStyle() & LVS_TYPEMASK};
  switch (dwStyle)
  {
    case LVS_ICON:
    {
      menu.CheckMenuItem(ID_SHELLCTRLS2_VIEW_LARGEICONS, MF_BYCOMMAND | MF_CHECKED);
      break;
    }
    case LVS_SMALLICON:
    {
      menu.CheckMenuItem(ID_SHELLCTRLS2_VIEW_SMALLICONS, MF_BYCOMMAND | MF_CHECKED);
      break;
    }
    case LVS_LIST:
    {
      menu.CheckMenuItem(ID_SHELLCTRLS2_VIEW_LIST, MF_BYCOMMAND | MF_CHECKED);
      break;
    }
    case LVS_REPORT:
    {
      menu.CheckMenuItem(ID_SHELLCTRLS2_VIEW_DETAILS, MF_BYCOMMAND | MF_CHECKED);
      break;
    }
    default:
    {
      break;
    }
  }
  const CMFCHeaderCtrl& headerCtrl{GetHeaderCtrl()};
  if (headerCtrl.GetSafeHwnd() == nullptr)
  {
    menu.EnableMenuItem(ID_SHELLCTRLS2_SORTBY_ASCENDING, MF_BYCOMMAND | MF_GRAYED);
    menu.EnableMenuItem(ID_SHELLCTRLS2_SORTBY_DESCENDING, MF_BYCOMMAND | MF_GRAYED);
    menu.EnableMenuItem(ID_SHELLCTRLS2_SORTBY_NAME, MF_BYCOMMAND | MF_GRAYED);
    menu.EnableMenuItem(ID_SHELLCTRLS2_SORTBY_SIZE, MF_BYCOMMAND | MF_GRAYED);
    menu.EnableMenuItem(ID_SHELLCTRLS2_SORTBY_TYPE, MF_BYCOMMAND | MF_GRAYED);
    menu.EnableMenuItem(ID_SHELLCTRLS2_SORTBY_DATEMODIFIED, MF_BYCOMMAND | MF_GRAYED);
  }
  else
  {
    const int nSortColumn{headerCtrl.GetSortColumn()};
    const int nColumnState{headerCtrl.GetColumnState(nSortColumn)};
    if (nColumnState == 1)
      menu.CheckMenuItem(ID_SHELLCTRLS2_SORTBY_ASCENDING, MF_BYCOMMAND | MF_CHECKED);
    else if (nColumnState == -1)
      menu.CheckMenuItem(ID_SHELLCTRLS2_SORTBY_DESCENDING, MF_BYCOMMAND | MF_CHECKED);
    switch (nSortColumn)
    {
      case static_cast<int>(Columns::NameColumn):
      {
        menu.CheckMenuItem(ID_SHELLCTRLS2_SORTBY_NAME, MF_BYCOMMAND | MF_CHECKED);
        break;
      }
      case static_cast<int>(Columns::DateModifiedColumn):
      {
        menu.CheckMenuItem(ID_SHELLCTRLS2_SORTBY_DATEMODIFIED, MF_BYCOMMAND | MF_CHECKED);
        break;
      }
      case static_cast<int>(Columns::TypeColumn):
      {
        menu.CheckMenuItem(ID_SHELLCTRLS2_SORTBY_TYPE, MF_BYCOMMAND | MF_CHECKED);
        break;
      }
      case static_cast<int>(Columns::SizeColumn):
      {
        menu.CheckMenuItem(ID_SHELLCTRLS2_SORTBY_SIZE, MF_BYCOMMAND | MF_CHECKED);
        break;
      }
      default:
      {
        break;
      }
    }
  }
  const CPoint point{GetCurrentMessage()->pt};
#pragma warning(suppress: 26472)
  const UINT nIDCmd{static_cast<UINT>(pPopup->TrackPopupMenu(TPM_LEFTALIGN | TPM_RETURNCMD | TPM_RIGHTBUTTON, point.x, point.y, this))};
  if (nIDCmd == 0)
    return;
  switch (nIDCmd)
  {
    case ID_SHELLCTRLS2_VIEW_LARGEICONS:
    {
      ModifyStyle(LVS_SMALLICON | LVS_LIST | LVS_REPORT, LVS_ICON);
      break;
    }
    case ID_SHELLCTRLS2_VIEW_SMALLICONS:
    {
      ModifyStyle(LVS_ICON | LVS_LIST | LVS_REPORT, LVS_SMALLICON);
      break;
    }
    case ID_SHELLCTRLS2_VIEW_LIST:
    {
      ModifyStyle(LVS_SMALLICON | LVS_REPORT | LVS_ICON, LVS_LIST);
      break;
    }
    case ID_SHELLCTRLS2_VIEW_DETAILS:
    {
      ModifyStyle(LVS_SMALLICON | LVS_LIST | LVS_ICON, LVS_REPORT);
      break;
    }
    case ID_SHELLCTRLS2_SORTBY_ASCENDING:
    {
      Sort(headerCtrl.GetSortColumn(), TRUE);
      break;
    }
    case ID_SHELLCTRLS2_SORTBY_DESCENDING:
    {
      Sort(headerCtrl.GetSortColumn(), FALSE);
      break;
    }
    case ID_SHELLCTRLS2_SORTBY_NAME:
    {
      const int nSortColumn{headerCtrl.GetSortColumn()};
      BOOL bSortAscending{TRUE};
      if (nSortColumn != -1)
        bSortAscending = headerCtrl.IsAscending();
      Sort(static_cast<int>(CShellListCtrl2::Columns::NameColumn), bSortAscending);
      break;
    }
    case ID_SHELLCTRLS2_SORTBY_SIZE:
    {
      const int nSortColumn{headerCtrl.GetSortColumn()};
      BOOL bSortAscending{TRUE};
      if (nSortColumn != -1)
        bSortAscending = headerCtrl.IsAscending();
      Sort(static_cast<int>(CShellListCtrl2::Columns::SizeColumn), bSortAscending);
      break;
    }
    case ID_SHELLCTRLS2_SORTBY_TYPE:
    {
      const int nSortColumn{headerCtrl.GetSortColumn()};
      BOOL bSortAscending{TRUE};
      if (nSortColumn != -1)
        bSortAscending = headerCtrl.IsAscending();
      Sort(static_cast<int>(CShellListCtrl2::Columns::TypeColumn), bSortAscending);
      break;
    }
    case ID_SHELLCTRLS2_SORTBY_DATEMODIFIED:
    {
      const int nSortColumn{headerCtrl.GetSortColumn()};
      BOOL bSortAscending{TRUE};
      if (nSortColumn != -1)
        bSortAscending = headerCtrl.IsAscending();
      Sort(static_cast<int>(CShellListCtrl2::Columns::DateModifiedColumn), bSortAscending);
      break;
    }
    case ID_SHELLCTRLS2_REFRESH:
    {
      Refresh();
      break;
    }
    default:
    {
      break;
    }
  }
}

void CShellListCtrl2::OnShowShellContextMenu(_In_ CPoint point)
{
  if (m_ContextMenu2 != nullptr)
    return;
  if (!m_bShellContextMenu)
    return;
  if (m_psfCurFolder == nullptr)
    return;

  const UINT nSelItems{GetSelectedCount()};
  int nClickedItem = -1;
  if ((point.x == -1) && (point.y == -1))
  {
    //Keyboard, show menu for the currently selected item(s)
    if (nSelItems == 0)
      return;
    int nCurItem{-1};
    int nLastSelItem{-1};
    for (UINT i{0}; i<nSelItems; i++)
    {
      nCurItem = GetNextItem(nCurItem, LVNI_SELECTED);
      nLastSelItem = nCurItem;
    }
    CRect rectItem;
    if (GetItemRect(nLastSelItem, rectItem, LVIR_BOUNDS))
    {
      point.x = rectItem.left;
      point.y = rectItem.bottom + 1;
      ClientToScreen(&point);
    }
  }
  else
  {
    //Clicked on specifed item
    LVHITTESTINFO lvhti{};
    lvhti.pt = point;
    ScreenToClient(&lvhti.pt);
    lvhti.flags = LVHT_NOWHERE;
    HitTest(&lvhti);
    if ((lvhti.flags & LVHT_ONITEM) == 0)
      return; //Click outside of items, do nothing
    nClickedItem = lvhti.iItem;
  }

  //Get the selected items
  std::vector<CPIDL> pidls{nSelItems, std::allocator<CPIDL>{}};
  LVITEM lvItem{};
  lvItem.mask = LVIF_PARAM;
  CShellItemInfo2* pClickedInfo{nullptr};
  if (nSelItems && (nClickedItem != -1))
  {
    //Put the item clicked on first in the list
    lvItem.iItem = nClickedItem;
    if (GetItem(&lvItem))
    {
#pragma warning(suppress: 26490)
      pClickedInfo = reinterpret_cast<CShellItemInfo2*>(lvItem.lParam);
      ASSERT(pClickedInfo != nullptr);

#pragma warning(suppress: 26446 26489)
      pidls[0] = pClickedInfo->m_pidlRel;
    }
  }

  int nCurItem{-1};
  for (UINT i{(nClickedItem >= 0) ? 1u : 0u}; i<nSelItems; i++)
  {
    nCurItem = GetNextItem(nCurItem, LVNI_SELECTED);
    if (nCurItem != nClickedItem)
    {
      lvItem.iItem = nCurItem;
      if (GetItem(&lvItem))
      {
#pragma warning(suppress: 26490)
        auto pInfo{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
        ASSERT(pInfo != nullptr);
#pragma warning(suppress: 26446 26489)
        pidls[i] = pInfo->m_pidlRel;
        if (pClickedInfo == nullptr)
          pClickedInfo = pInfo;
      }
    }
    else
      i--;
  }

  ATL::CComPtr<IContextMenu> pcm;
#pragma warning(suppress: 26490)
  if (FAILED(m_psfCurFolder->GetUIObjectOf(GetSafeHwnd(), nSelItems, reinterpret_cast<LPCITEMIDLIST*>(pidls.data()), IID_IContextMenu, nullptr, reinterpret_cast<LPVOID*>(&pcm))))
    return;

  CMenu popup;
  if (!popup.CreatePopupMenu())
    return;
  if (FAILED(pcm->QueryContextMenu(popup.m_hMenu, 0, 1, 0x7FFF, CMF_NORMAL | CMF_EXPLORE)))
    return;
#pragma warning(suppress: 26490)
  if (FAILED(pcm->QueryInterface(IID_IContextMenu2, reinterpret_cast<LPVOID*>(&m_ContextMenu2))))
    return;
#pragma warning(suppress: 26472)
  const UINT nIDCmd{static_cast<UINT>(popup.TrackPopupMenu(TPM_LEFTALIGN | TPM_RETURNCMD | TPM_RIGHTBUTTON, point.x, point.y, this))};
  if (m_ContextMenu2 != nullptr)
    m_ContextMenu2.Release();
  if (nIDCmd == 0)
    return;

  if ((pClickedInfo != nullptr) && (nSelItems == 1) && (nIDCmd == popup.GetDefaultItem(0, FALSE)))
  {
    //If specified element is a folder, try to display it
    SFGAOF attribs{SFGAO_FOLDER};
    m_psfCurFolder->GetAttributesOf(1, pClickedInfo->m_pidlRel, &attribs);
    if (attribs & SFGAO_FOLDER)
    {
#pragma warning(suppress: 26486)
      DisplayFolder(pClickedInfo);
      return;
    }
  }

  const CWnd* pParent{GetParent()};
  if (pParent == nullptr)
  {
    ASSERT(FALSE);
    return;
  }
  CMINVOKECOMMANDINFO cmi{};
  cmi.cbSize = sizeof(CMINVOKECOMMANDINFO);
  cmi.hwnd = pParent->GetSafeHwnd();
#pragma warning(suppress: 26472 26490)
  cmi.lpVerb = reinterpret_cast<LPCSTR>(static_cast<INT_PTR>(nIDCmd - 1));
  cmi.nShow = SW_SHOWNORMAL;
  const HRESULT hr{pcm->InvokeCommand(&cmi)};
  if (SUCCEEDED(hr))
#pragma warning(suppress: 26472)
    pParent->SendMessage(WM_SHELLLISTCTRL2_AFTER_SHELL_COMMAND, static_cast<WPARAM>(nIDCmd));
  SetFocus();
  if (m_bRefreshAfterShellCommand)
    Refresh();
}

#pragma warning(suppress: 26434)
void CShellListCtrl2::OnDestroy()
{
  KillMonitoringThread(); //NOLINT(clang-analyzer-cplusplus.NewDelete)
  CShellTreeCtrl2* pTree{GetRelatedTree()};
  if (pTree != nullptr)
    pTree->m_hRelatedList = nullptr;
  __super::OnDestroy();
}

CShellTreeCtrl2* CShellListCtrl2::GetRelatedTree() const
{
  if ((m_hRelatedTree == nullptr) || !::IsWindow(m_hRelatedTree))
    return nullptr;
  return DYNAMIC_DOWNCAST(CShellTreeCtrl2, CWnd::FromHandlePermanent(m_hRelatedTree));
}

void CShellListCtrl2::SetRelatedTree(_In_opt_ CShellTreeCtrl2* pShellTree)
{
  m_hRelatedTree = (pShellTree == nullptr) ? nullptr : pShellTree->GetSafeHwnd();
  if (pShellTree != nullptr)
    pShellTree->m_hRelatedList = GetSafeHwnd();
}

void CShellListCtrl2::SetEnumObjectFlags(_In_ SHCONTF flags, _In_ bool bRefresh)
{
  if (flags != m_EnumObjectFlags)
  {
    m_EnumObjectFlags = flags;
    if (m_EnumObjectFlags & SHCONTF_NONFOLDERS)
      m_bShowFiles = true;
    else
      m_bShowFiles = false;
    if (bRefresh && (GetSafeHwnd() != nullptr))
      Refresh();
  }
}

LRESULT CShellListCtrl2::WindowProc(UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
    case WM_INITMENUPOPUP: //deliberate fallthrough
    case WM_DRAWITEM: //deliberate fallthrough
    case WM_MEASUREITEM:
    {
      if (m_ContextMenu2 != nullptr)
      {
        m_ContextMenu2->HandleMenuMsg(message, wParam, lParam);
        return 0;
      }
      break;
    }
    default:
    {
      break;
    }
  }
  return __super::WindowProc(message, wParam, lParam);
}

#pragma warning(suppress: 26440)
void CShellListCtrl2::OnItemAdded(_In_ int /*nItem*/)
{
  //Derived classes are free to do work here
}

#pragma warning(suppress: 26434)
DWORD_PTR CShellListCtrl2::GetItemData(_In_ int nItem) const
{
#pragma warning(suppress: 26490)
  const CShellItemInfo2* pItem{reinterpret_cast<CShellItemInfo2*>(__super::GetItemData(nItem))};
  ASSERT(pItem != nullptr);
#pragma warning(suppress: 26489)
  return pItem->m_dwItemData;
}

#pragma warning(suppress: 26434)
_Success_(return != false) bool CShellListCtrl2::SetItemData(_In_ int nItem, _In_ DWORD_PTR dwData)
{
  LVITEM lvItem{};
  lvItem.mask = LVIF_PARAM;
  lvItem.iItem = nItem;
  if (!GetItem(&lvItem))
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26490)
  auto pItem{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
  if (pItem == nullptr)
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26489)
  pItem->m_dwItemData = dwData;
  return true;
}

void* CShellListCtrl2::GetItemDataPtr(_In_ int nItem) const
{
#pragma warning(suppress: 26490)
  const CShellItemInfo2* pItem{reinterpret_cast<CShellItemInfo2*>(__super::GetItemData(nItem))};
  ASSERT(pItem != nullptr);
#pragma warning(suppress: 26489 26490)
  return reinterpret_cast<void*>(pItem->m_dwItemData);
}

#pragma warning(suppress: 26434)
_Success_(return != false) bool CShellListCtrl2::SetItemDataPtr(_In_ int nItem, _In_opt_ void* pData)
{
  LVITEM lvItem{};
  lvItem.mask = LVIF_PARAM;
  lvItem.iItem = nItem;
  if (!GetItem(&lvItem))
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26490)
  auto pItem{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
  if (pItem == nullptr)
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26489 26490)
  pItem->m_dwItemData = reinterpret_cast<DWORD_PTR>(pData);
  return true;
}

void CShellListCtrl2::SetShowFiles(_In_ bool bShowFiles, _In_ bool bRefresh)
{
  if (bShowFiles != m_bShowFiles)
  {
    m_bShowFiles = bShowFiles;
    if (bShowFiles)
      m_EnumObjectFlags |= SHCONTF_NONFOLDERS;
    else
      m_EnumObjectFlags &= ~SHCONTF_NONFOLDERS;
    if (bRefresh && (GetSafeHwnd() != nullptr))
      Refresh();
  }
}

_Success_(return != false) bool CShellListCtrl2::DoShellProperties(_In_ int nItem)
{
  LVITEM lvItem{};
  lvItem.mask = LVIF_PARAM;
  lvItem.iItem = nItem;
  if (!GetItem(&lvItem))
    return false;
  //Execute the "properties" command for the selected item
#pragma warning(suppress: 26490)
  const CShellItemInfo2* pItem{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
  if (pItem == nullptr)
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26476)
  SHELLEXECUTEINFO sei{};
  sei.cbSize = sizeof(sei);
  const CWnd* pParent{GetParent()};
  if (pParent == nullptr)
  {
    ASSERT(FALSE);
    return false;
  }
  sei.hwnd = pParent->GetSafeHwnd();
  sei.nShow = SW_SHOW;
#pragma warning(suppress: 26489)
  sei.lpIDList = pItem->m_pidlFQ.m_pPIDL;
  sei.lpVerb = _T("properties");
  sei.fMask = SEE_MASK_INVOKEIDLIST | SEE_MASK_IDLIST;
#pragma warning(suppress: 26486)
  return ShellExecuteEx(&sei);
}

void CShellListCtrl2::SetDriveHideFlags(_In_ DWORD dwDriveHideFlags, _In_ bool bRefresh)
{
  m_dwDriveHideFlags = dwDriveHideFlags;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::SetFileHideFlags(_In_ DWORD dwFileHideFlags, _In_ bool bRefresh)
{
  m_dwFileHideFlags = dwFileHideFlags;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::SetFolderHideFlags(_In_ DWORD dwFolderHideFlags, _In_ bool bRefresh)
{
  m_dwFolderHideFlags = dwFolderHideFlags;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

size_t CShellListCtrl2::AddMultipleFilenameMasks(_In_ const CString& sFilenameMasks, _In_ bool bRefresh)
{
  //Empty out the array
  m_FilenameMasks.clear();

  int nCurPos{0};
  CString sToken{sFilenameMasks.Tokenize(_T(",;"), nCurPos)};
  while (sToken != _T("")) //NOLINT(clang-analyzer-core.CallAndMessage)
  {
    sToken.Trim();
    m_FilenameMasks.push_back(sToken);
    sToken = sFilenameMasks.Tokenize(_T(",;"), nCurPos);
  }

  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();

  //Return the number of recipients parsed
  return m_FilenameMasks.size();
}

CString CShellListCtrl2::GetFilenameMasks() const
{
  //What will be the return value from this function
  CString sMasks;

  const auto nSize{m_FilenameMasks.size()};
  for (size_t i{0}; i<nSize; i++)
  {
#pragma warning(suppress: 26446)
    sMasks += m_FilenameMasks[i];
    if (i < (nSize - 1))
      sMasks += _T(",");
  }
  return sMasks;
}

void CShellListCtrl2::SetFilenameMask(_In_ size_t nIndex, _In_ const CString& sFileNameMask, _In_ bool bRefresh)
{
#pragma warning(suppress: 26446)
  m_FilenameMasks[nIndex] = sFileNameMask;

  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::AddFilenameMask(_In_ const CString& sFileNameMask, _In_ bool bRefresh)
{
  m_FilenameMasks.push_back(sFileNameMask);
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::SetCompressedColor(_In_ COLORREF rgbCompressed, _In_ bool bRefresh)
{
  m_rgbCompressed = rgbCompressed;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::SetEncryptedColor(_In_ COLORREF rgbEncrypted, _In_ bool bRefresh)
{
  m_rgbEncrypted = rgbEncrypted;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::SetUsingDifferentColorForCompressed(_In_ bool bShowCompressedUsingDifferentColor, _In_ bool bRefresh)
{
  m_bShowCompressedUsingDifferentColor = bShowCompressedUsingDifferentColor;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::SetUsingDifferentIconForSharedFolders(_In_ bool bShowSharedUsingDifferentIcon, _In_ bool bRefresh)
{
  m_bShowSharedUsingDifferentIcon = bShowSharedUsingDifferentIcon;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::SetUsingDifferentIconForLinkedItems(_In_ bool bShowLinkedItemsUsingDifferentIcon, _In_ bool bRefresh)
{
  m_bShowLinkedItemsUsingDifferentIcon = bShowLinkedItemsUsingDifferentIcon;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
}

void CShellListCtrl2::SetUsingDifferentColorForEncrypted(_In_ bool bShowEncryptedUsingDifferentColor, _In_ bool bRefresh)
{
  m_bShowEncryptedUsingDifferentColor = bShowEncryptedUsingDifferentColor;
  if (bRefresh && (GetSafeHwnd() != nullptr))
    Refresh();
};

_Success_(return != false) bool CShellListCtrl2::CanDisplayFile(_In_z_ LPCTSTR pszFilename)
{
  //Now check to see if there is any files of the specified file mask
  std::filesystem::path path{pszFilename};
  for (const auto& sMask : m_FilenameMasks)
  {
    std::filesystem::path mask{sMask.GetString()};
    path.replace_filename(mask.filename());
    path.replace_extension(mask.extension());
    CFileFind find;
#ifdef _UNICODE
    BOOL bFind{find.FindFile(path.c_str())};
#else
    BOOL bFind{find.FindFile(CStringA(path.c_str()))};
#endif //#ifdef _UNICODE
    while (bFind)
    {
      bFind = find.FindNextFile();
      if ((find.GetFilePath().CompareNoCase(pszFilename) == 0) && !find.MatchesMask(m_dwFileHideFlags))
        return true;
    }
  }
  return false;
}

#pragma warning(suppress: 26440)
_Success_(return != false) bool CShellListCtrl2::CanDisplayFolder(_In_z_ LPCTSTR pszFolder)
{
  //Derived classes can decide dynamically whether or not a certain folder is to be displayed. CShellListCtrl2 by default
  //displays all folders which do not have attributes as set in m_dwFolderHideFlags

  const DWORD dwAttributes{GetFileAttributes(pszFolder)};
  return ((dwAttributes != INVALID_FILE_ATTRIBUTES) && ((dwAttributes & m_dwFolderHideFlags) == 0));
}

#pragma warning(suppress: 26440)
_Success_(return != false) bool CShellListCtrl2::CanDisplayDrive(_In_ const CString& sDrive)
{
  //Derived classes can decide dynamically whether or not a certain drive is to be displayed.
  //CShellListCtrl2 by default displays all drives which do not have attributes as set in m_dwDriveHideFlags

  //check if this drive is one of the types to hide
  bool bDisplay{true};
  const UINT nDrive{GetDriveType(sDrive)};
  switch (nDrive)
  {
    case DRIVE_REMOVABLE:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_REMOVABLE))
        bDisplay = false;
      break;
    }
    case DRIVE_FIXED:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_FIXED))
        bDisplay = false;
      break;
    }
    case DRIVE_REMOTE:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_REMOTE))
        bDisplay = false;
      break;
    }
    case DRIVE_CDROM:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_CDROM))
        bDisplay = false;
      break;
    }
    case DRIVE_RAMDISK:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_RAMDISK))
        bDisplay = false;
      break;
    }
    default:
    {
      break;
    }
  }

  return bDisplay;
}

_Success_(return != false) bool CShellListCtrl2::CanDisplay(_In_ const CPIDL& pidlParent, _In_ const CPIDL& pidl)
{
  //Derived classes can decide dynamically whether or not a certain PIDL is to be displayed.
  //The default implementation checks via CanDisplayFolder for file system folders,
  //CanDisplayFile for file system files (non-folders) and CanDisplayDrive for drive letters
  //and if not any of the above, then allows the display

  bool bAddItem{false};
  CPIDL fqPidl{pidlParent + pidl};
  CString sPath;
  if (fqPidl.ToPath(sPath))
  {
    const DWORD dwAttributes{GetFileAttributes(sPath)};
    const bool bIsFile{(dwAttributes != INVALID_FILE_ATTRIBUTES) && ((dwAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)};
    const bool bIsFolder{(dwAttributes != INVALID_FILE_ATTRIBUTES) && (dwAttributes & FILE_ATTRIBUTE_DIRECTORY)};
#pragma warning(suppress: 26446)
    if ((m_FilenameMasks.size() == 1) && (m_FilenameMasks[0] == _T("*.*")))
    {
      if (IsDrive(sPath))
        bAddItem = CanDisplayDrive(sPath);
      else if (bIsFolder)
        bAddItem = ((dwAttributes & m_dwFolderHideFlags) == 0);
      else if (bIsFile)
        bAddItem = ((dwAttributes & m_dwFileHideFlags) == 0);
      else
        bAddItem = false;
    }
    else
    {
      if (IsDrive(sPath))
        bAddItem = CanDisplayDrive(sPath);
      else if (bIsFolder)
        bAddItem = CanDisplayFolder(sPath);
      else if (bIsFile)
        bAddItem = CanDisplayFile(sPath);
      else
        bAddItem = false;
    }
  }
  else
    bAddItem = true;

  return bAddItem;
}

#pragma warning(suppress: 26429 26434)
void CShellListCtrl2::OnCustomDraw(NMHDR* pNMHDR, LRESULT* pResult)
{
  //Validate our parameters
  ASSERT(pResult != nullptr);

#pragma warning(suppress: 26490)
  auto pCustomDraw{reinterpret_cast<NMLVCUSTOMDRAW*>(pNMHDR)};
  ASSERT(pCustomDraw != nullptr);
  switch (pCustomDraw->nmcd.dwDrawStage)
  {
    case CDDS_PREPAINT:
    {
      *pResult = CDRF_NOTIFYITEMDRAW; //Tell the control that we are interested in item notifications
      break;
    }
    case CDDS_ITEMPREPAINT:
    {
      //Let it know that we are finished doing custom draw
      *pResult = CDRF_DODEFAULT;

      //To work around what looks like a bug in CommCtrl v6 we do not do custom drawing until the window is visible
      if (!IsWindowVisible())
        return;

  #pragma warning(suppress: 26472 26490)
      if (m_bShowCompressedUsingDifferentColor && IsCompressed(static_cast<int>(pCustomDraw->nmcd.dwItemSpec)))
        pCustomDraw->clrText = m_rgbCompressed;
  #pragma warning(suppress: 26472 26490)
      else if (m_bShowEncryptedUsingDifferentColor && IsEncrypted(static_cast<int>(pCustomDraw->nmcd.dwItemSpec)))
        pCustomDraw->clrText = m_rgbEncrypted;

      break;
    }
    default:
    {
      break;
    }
  }
}

BOOL CShellListCtrl2::PreTranslateMessage(MSG* pMsg)
{
  //When an item is being edited make sure the edit control
  //receives certain important key strokes
  if (GetEditControl())
  {
    ::TranslateMessage(pMsg);
    ::DispatchMessage(pMsg);
    return TRUE; //DO NOT process further
  }

  //Context menu via the keyboard
  if (((((pMsg->message == WM_KEYDOWN) || (pMsg->message == WM_SYSKEYDOWN)) &&
      (pMsg->wParam == VK_F10) && (GetKeyState(VK_SHIFT) & 0x8000)) != 0) ||
      (pMsg->message == WM_CONTEXTMENU))
  {
    const CPoint p{GetCurrentMessage()->pt};
    OnContextMenu(nullptr, p);
    return TRUE;
  }
  //Hitting the Enter key opens the item
  else if (pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_RETURN)
  {
    const int nItem{GetNextItem(-1, LVNI_FOCUSED)};
    if (nItem != -1)
    {
      DoShellDefault(nItem);
      return TRUE;
    }
  }
  //hitting the F5 key forces a refresh of the whole tree
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_F5))
  {
    Refresh();
    return TRUE;
  }
  //Hitting the backspace shows the parent folder
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_BACK))
  {
    DisplayParentFolder();
    return TRUE;
  }

  //Let the parent class do its thing
  return __super::PreTranslateMessage(pMsg);
}

_Success_(return != false) bool CShellListCtrl2::IsCompressed(_In_ int nItem)
{
  LVITEM lvItem{};
  lvItem.iItem = nItem;
  lvItem.mask = TVIF_PARAM;
  if (!GetItem(&lvItem))
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26490)
  const CShellItemInfo2* pInfo{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
  if (pInfo == nullptr)
  {
    ASSERT(FALSE);
    return false;
  }
  ASSERT(pInfo != nullptr);
#pragma warning(suppress: 26489)
  return (pInfo->m_Attribs & SFGAO_COMPRESSED);
}

_Success_(return != false) bool CShellListCtrl2::IsEncrypted(_In_ int nItem)
{
  LVITEM lvItem{};
  lvItem.iItem = nItem;
  lvItem.mask = TVIF_PARAM;
  if (!GetItem(&lvItem))
  {
    ASSERT(FALSE);
    return false;
  }
#pragma warning(suppress: 26490)
  const CShellItemInfo2* pInfo{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
  if (pInfo == nullptr)
  {
    ASSERT(FALSE);
    return false;
  }

#pragma warning(suppress: 26489)
  return (pInfo->m_Attribs & SFGAO_ENCRYPTED);
}

#pragma warning(suppress: 26429)
_Success_(return != false) bool CShellListCtrl2::IsDrive(_In_z_ LPCTSTR pszPath) noexcept
{
#pragma warning(suppress: 26481)
  return (_tcslen(pszPath) == 3) && (pszPath[1] == _T(':')) && (pszPath[2] == _T('\\'));
}

void CShellListCtrl2::CreateMonitoringThread(_In_z_ LPCTSTR pszPath)
{
  //Kill any current notification thread
  KillMonitoringThread();

  //Setup the monitoring path
  m_sMonitoringPath = pszPath;
  ASSERT(m_pMonitoringThread == nullptr);
  m_pMonitoringThread = AfxBeginThread(_MonitoringThread, this, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED);
  if (m_pMonitoringThread != nullptr)
  {
    //We're in charge of deletion of the thread
    m_pMonitoringThread->m_bAutoDelete = FALSE;

    //Resume the thread now that everything is ready to go
    m_pMonitoringThread->ResumeThread();
  }
}

UINT CShellListCtrl2::_MonitoringThread(_In_ LPVOID pParam)
{
  //Validate our parameters
  ASSERT(pParam != nullptr);
  auto pThis{static_cast<CShellListCtrl2*>(pParam)};
#pragma warning(suppress: 26496)
  AFXASSUME(pThis != nullptr);

  //Convert from the SDK world to the C++ world
  return pThis->MonitoringThread(); //NOLINT(clang-analyzer-core.CallAndMessage)
}

UINT CShellListCtrl2::MonitoringThread()
{
  //Form the notification flag to use
  DWORD dwNotifyFilter{FILE_NOTIFY_CHANGE_DIR_NAME};
  if (m_bShowFiles)
    dwNotifyFilter |= (FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_SIZE | FILE_NOTIFY_CHANGE_LAST_WRITE);

  //Get a handle to a file change notification object
  HANDLE hChange{FindFirstChangeNotification(m_sMonitoringPath, TRUE, dwNotifyFilter)};
  if (hChange != INVALID_HANDLE_VALUE)
  {
    //Setup the handles to wait on
    std::array<HANDLE, 2> handles{ { hChange, m_MonitoringTerminateEvent } };

    //Sleep until a file change notification wakes this thread or m_TerminateEvent becomes
    //set indicating it's time for the thread to end
    bool bContinue{true};
    while (bContinue)
    {
#pragma warning(suppress: 26472)
      const DWORD dwWait{WaitForMultipleObjects(static_cast<DWORD>(handles.size()), handles.data(), FALSE, INFINITE)};
      if ((dwWait - WAIT_OBJECT_0) == 0)
      {
        //Respond to the change notification by posting a user defined message back to the GUI thread
        if (!m_bAutoRefresh)
          bContinue = false;
        else
          PostMessage(WM_SHELLLISTCTRL2_NOTIFICATION_CHANGE);

        //Move onto the next notification
        FindNextChangeNotification(hChange);
      }
      else
      {
        //Kill the thread
        bContinue = false;
      }
    }

    //Close the handle we have open
    FindCloseChangeNotification(hChange);
  }

  return 0;
}

void CShellListCtrl2::KillMonitoringThread()
{
  if (m_pMonitoringThread != nullptr)
  {
    //Signal the worker thread to exit and wait for it to return
    m_MonitoringTerminateEvent.SetEvent();
    WaitForSingleObject(m_pMonitoringThread->m_hThread, INFINITE);
    delete m_pMonitoringThread;
    m_pMonitoringThread = nullptr;
  }
}

LRESULT CShellListCtrl2::OnNotificationChange(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  if (m_bAutoRefresh)
    Refresh();
  return 0L;
}

_Success_(return != false) bool CShellListCtrl2::CanHandleChangeNotifications(_In_z_ LPCTSTR pszPath)
{
  //The path must be a folder of some kind
  const DWORD dwAttributes{GetFileAttributes(pszPath)};
  if (dwAttributes == INVALID_FILE_ATTRIBUTES)
    return false;
  if ((dwAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
    return false;

  //check if this drive is one of the types which can issue notification changes
  std::filesystem::path path{pszPath};
  const UINT nDrive{GetDriveTypeW(path.root_path().c_str())};
  return ((nDrive != DRIVE_REMOVABLE) && (nDrive != DRIVE_CDROM));
}

void CShellListCtrl2::SetAutoRefresh(_In_ bool bAutoRefresh)
{
  if (bAutoRefresh != m_bAutoRefresh) //NOLINT(clang-analyzer-core.CallAndMessage)
  {
    m_bAutoRefresh = bAutoRefresh;

    //Force the monitoring thread to be recreated
    Refresh();
  }
}

#pragma warning(suppress: 26429)
void CShellListCtrl2::OnBeginDrag(NMHDR* /*pNMHDR*/, LRESULT* pResult)
{
  *pResult = 0;
  if (!m_bAllowDragDrop) //NOLINT(clang-analyzer-core.CallAndMessage)
    return;

  //Get the selected items into a array of LPCITEMIDLIST
  //Also work out the dwEffect value for the combined set of items
  DROPEFFECT dwEffect{DROPEFFECT_COPY | DROPEFFECT_MOVE | DROPEFFECT_LINK};
  const int nItemCount{GetItemCount()};
  std::vector<LPCITEMIDLIST> selPIDLS;
  selPIDLS.reserve(nItemCount);
  POSITION posItem{GetFirstSelectedItemPosition()};
  while (posItem != nullptr)
  {
    const int nSelItem{GetNextSelectedItem(posItem)};
    if (!IsDropSource(nSelItem))
      return;
    LVITEM lvItem{};
    lvItem.iItem = nSelItem;
    lvItem.mask = LVIF_PARAM;
    if (!GetItem(&lvItem))
    {
      ASSERT(FALSE);
      return;
    }
#pragma warning(suppress: 26490)
    const CShellItemInfo2* pItem{reinterpret_cast<CShellItemInfo2*>(lvItem.lParam)};
    if (pItem == nullptr)
    {
      ASSERT(FALSE);
      return;
    }
    ASSERT(pItem->m_pParentFolder != nullptr); //Validate our parameters
    SFGAOF attribs{SFGAO_CANCOPY | SFGAO_CANMOVE | SFGAO_CANLINK};
    const HRESULT hr{pItem->m_pParentFolder->GetAttributesOf(1, pItem->m_pidlFQ, &attribs)};
    if (FAILED(hr))
    {
      ASSERT(FALSE);
      return;
    }
    if ((attribs & SFGAO_CANCOPY) == 0)
      dwEffect &= ~DROPEFFECT_COPY;
    if ((attribs & SFGAO_CANMOVE) == 0)
      dwEffect &= ~DROPEFFECT_MOVE;
    if ((attribs & SFGAO_CANLINK) == 0)
      dwEffect &= ~DROPEFFECT_LINK;
    selPIDLS.push_back(pItem->m_pidlFQ);
  }
  if (dwEffect == 0)
    return;
  //Create the IShellItemArray COM ptr
  ATL::CComPtr<IShellItemArray> pShellItems;
#pragma warning(suppress: 26472)
  HRESULT hr{SHCreateShellItemArrayFromIDLists(static_cast<UINT>(selPIDLS.size()), selPIDLS.data(), &pShellItems)};
  if (FAILED(hr))
  {
    ASSERT(FALSE);
    return;
  }
  ATL::CComPtr<IDataObject> pDataObject;
#pragma warning(suppress: 26490)
  hr = pShellItems->BindToHandler(nullptr, BHID_DataObject, __uuidof(pDataObject), reinterpret_cast<void**>(&pDataObject));
  if (FAILED(hr))
  {
    ASSERT(FALSE);
    return;
  }
  DROPEFFECT deResult{DROPEFFECT_NONE};
  hr = SHDoDragDrop(nullptr, pDataObject, nullptr, dwEffect, &deResult);
  if (FAILED(hr))
  {
    ASSERT(FALSE);
    return;
  }
}

void CShellListCtrl2::OnBeginRDrag(NMHDR* pNMHDR, LRESULT* pResult)
{
  OnBeginDrag(pNMHDR, pResult);
}

#pragma warning(suppress: 26440)
_Success_(return != false) bool CShellListCtrl2::IsDropSource(_In_ int /*nItem*/)
{
  //Derived classes can decide dynamically whether or not a certain node can be dragged.
  return true;
}

#pragma warning(suppress: 26429 26440)
DROPEFFECT CShellListCtrl2::OnDragEnter(_In_ COleDataObject* pDataObject, _In_ DWORD dwKeyState, _In_ CPoint point)
{
  //Validate our parameters
  ASSERT(pDataObject != nullptr);

  DROPEFFECT dwEffect{DROPEFFECT_NONE};
  HGLOBAL hGlobal{pDataObject->GetGlobalData(CF_HDROP)};
  if (hGlobal != nullptr)
  {
    m_bCanAcceptData = true;
    dwEffect = (dwKeyState & MK_CONTROL) ? DROPEFFECT_COPY : DROPEFFECT_MOVE;
    GlobalFree(hGlobal);
  }
  else
  {
    m_bCanAcceptData = false;
    dwEffect = DROPEFFECT_NONE;
  }
  if (dwKeyState & MK_RBUTTON)
    m_bLeftMouseDrag = false;
  else
    m_bLeftMouseDrag = true;
  if (!IsOverValidDropItem(point))
    dwEffect = DROPEFFECT_NONE;

  //Cache the mouse hover time
  UINT nMouseHoverTime{0};
  if (SystemParametersInfo(SPI_GETMOUSEHOVERTIME, 0, &nMouseHoverTime, 0))
    m_nMouseHoverTime = nMouseHoverTime;

  return dwEffect;
}

#pragma warning(suppress: 26440)
DROPEFFECT CShellListCtrl2::OnDragOver(_In_ COleDataObject* pDataObject, _In_ DWORD dwKeyState, _In_ CPoint point)
{
  //Validate our parameters
  ASSERT(pDataObject != nullptr);

  if (!m_bCanAcceptData)
    return DROPEFFECT_NONE;

  //check for hovering
  bool bHovering{false};
  const ULONGLONG dwCurrentTickCount{GetTickCount64()};
  if (point != m_lastMouseLoc)
  {
    //We are moving, so update the last mouse location
    m_lastMouseLoc = point;
    m_nLastMovementTime = dwCurrentTickCount;
  }
  else
  {
    //We are hovering (i.e. cursor hasn't moved) if the point value has not changed for "m_nMouseHoverTime" ms
    if ((dwCurrentTickCount - m_nLastMovementTime) > m_nMouseHoverTime)
      bHovering = true;
  }

  //check for mouse near top or bottom, scroll if necessary
  const int nFirstItem{GetTopIndex()};
  CRect ItemRect;
  GetItemRect(nFirstItem, &ItemRect, FALSE);
  CRect clientRect;
  GetClientRect(&clientRect);
  if (GetStyle() & WS_VSCROLL)
  {
    //Account for the possible header control at the top of the client area
    const CMFCHeaderCtrl& headerCtrl{GetHeaderCtrl()};
    if (headerCtrl.GetSafeHwnd() != nullptr)
    {
      CRect hdrRect;
      headerCtrl.GetWindowRect(hdrRect);
      ScreenToClient(hdrRect);
      clientRect.top += hdrRect.Height();
    }
    const int nHitDepth{ItemRect.Height() * 2};
    if ((point.y < (clientRect.top + nHitDepth)) && (point.y > clientRect.top))
    {
      //we need to scroll up
      if ((dwCurrentTickCount - m_nLastScrollTime) > 100)
      {
        SendMessage(WM_VSCROLL, SB_LINEUP);
        ResetHoverTimer(dwCurrentTickCount);
      }
    }
    else if (point.y > (clientRect.bottom - nHitDepth) && (point.y < clientRect.bottom))
    {
      //we need to scroll down
      if ((dwCurrentTickCount - m_nLastScrollTime) > 100)
      {
        SendMessage(WM_VSCROLL, SB_LINEDOWN);
        ResetHoverTimer(dwCurrentTickCount);
      }
    }
  }

  //Only select the drag item if location is in the client area. Also calculate the "dwEffect" value
  DWORD dwEffect{DROPEFFECT_NONE};
  if (clientRect.PtInRect(point))
  {
    UINT nFlags{0};
    const int nItem{HitTest(point, &nFlags) };
    CString sToPath;
    if (nItem != -1)
      GetItemPath(nItem, sToPath);
    else
      GetCurrentFolder(sToPath);

    CString sToVolume{CShellCtrl2DropTarget::GetVolumePathName(sToPath)};
    CString sFromPath;
    CShellCtrl2DropTarget::GetFirstPathFromDataObject(pDataObject, sFromPath);
    CString sFromVolume{CShellCtrl2DropTarget::GetVolumePathName(sFromPath)};
    if (!sToVolume.IsEmpty() && (sToVolume == sFromVolume))
      dwEffect = DROPEFFECT_MOVE;
    else
      dwEffect = DROPEFFECT_COPY;
    if (bHovering)
    {
      m_lastMouseLoc = point;
      m_nLastMovementTime = dwCurrentTickCount;
    }
    if (nItem != m_nItemDrop)
    {
      SelectDropTarget(nItem);
      m_nItemDrop = nItem;
    }
    //check for qualifier keys held down
    if (dwKeyState & MK_CONTROL) //user wants to force a copy operation
      dwEffect = DROPEFFECT_COPY;
    if ((dwEffect == DROPEFFECT_COPY) && (dwKeyState & MK_SHIFT)) //user wants to force a move operation
      dwEffect = DROPEFFECT_MOVE;
  }
  //check for hit on a valid item
  if ((dwEffect != DROPEFFECT_NONE) && !IsOverValidDropItem(point))
    dwEffect = DROPEFFECT_NONE;

  return dwEffect;
}

#pragma warning(suppress: 26440)
_Success_(return != FALSE) BOOL CShellListCtrl2::OnDrop(_In_ COleDataObject* /*pDataObject*/, _In_ DROPEFFECT /*dropEffect*/, _In_ CPoint /*point*/)
{
  //CShellListCtrl2 uses OnDropEx
  return FALSE;
}

#pragma warning(suppress: 26429 26440)
DROPEFFECT CShellListCtrl2::OnDropEx(_In_ COleDataObject* pDataObject, _In_ DROPEFFECT dropDefault, _In_ DROPEFFECT dropList, _In_ CPoint /*point*/)
{
  //Validate our parameters
  ASSERT(pDataObject != nullptr);

  DROPEFFECT deResult{DROPEFFECT_NONE};
  if (m_bCanAcceptData) //NOLINT(clang-analyzer-core.CallAndMessage)
  {
    HGLOBAL hGlobal{pDataObject->GetGlobalData(CF_HDROP)};
    if (hGlobal != nullptr)
    {
      CString sToPath;
      if (m_nItemDrop != -1)
        GetItemPath(m_nItemDrop, sToPath);
      else
        GetCurrentFolder(sToPath);
      HDROP hDrop{static_cast<HDROP>(GlobalLock(hGlobal))};
      if (hDrop != nullptr)
      {
        const bool bOldAutoRefresh(GetAutoRefresh());
        SetAutoRefresh(false); //Prevents us from getting thread notifications while we do the filesystem manipulation
        if (!m_bLeftMouseDrag)
        {
          //perform the actual filesystem manipulation after bringing up a context menu of choices
          deResult = PerformRightClickDragDrop(m_nItemDrop, dropDefault, dropList, hDrop, sToPath);
          SelectDropTarget(-1); //Remove the drop target in case PerformRightClickDragDrop above didn't do it
        }
        else
        {
          //perform the actual filesystem manipulation
          if (SUCCEEDED(PerformDragDrop(m_nItemDrop, dropDefault, hDrop, sToPath)))
            deResult = dropDefault;
        }
        SetAutoRefresh(bOldAutoRefresh);
        GlobalUnlock(hGlobal);
      }
      //Do a refresh if necessary depending on what was dropped and the "m_bShowFiles" setting
      if (deResult != DROPEFFECT_NONE)
      {
        bool bRefresh{false};
        if (m_bShowFiles)
          bRefresh = m_bDropContainedFiles || m_bDropContainedDirectories;
        else
          bRefresh = m_bDropContainedDirectories;
        if (bRefresh)
          Refresh();
      }
      GlobalFree(hGlobal);
    }
  }
  m_bCanAcceptData = false;
  return deResult;
}

#pragma warning(suppress: 26440)
void CShellListCtrl2::OnDragLeave()
{
  m_bCanAcceptData = false;
  SelectDropTarget(-1);
}

#pragma warning(suppress: 26440)
DROPEFFECT CShellListCtrl2::OnDragScroll(_In_ DWORD /*dwKeyState*/, _In_ CPoint /*point*/)
{
  return DROPEFFECT_NONE;
}

_Success_(return != false) bool CShellListCtrl2::IsOverValidDropItem(_In_ const CPoint& point)
{
  //check for hit on a valid tree item
  UINT flags{0};
  const int nItem{HitTest(point, &flags)};
  if (nItem == -1)
    return true;

  //The default behaviour is to only allow a drag over a folder in the tree control
  CString sPath;
  if (!GetItemPath(nItem, sPath))
    return false;
  const DWORD dwAttributes{GetFileAttributes(sPath)};
  return (dwAttributes != INVALID_FILE_ATTRIBUTES) && (dwAttributes & FILE_ATTRIBUTE_DIRECTORY);
}

#pragma warning(suppress: 26440)
void CShellListCtrl2::ResetHoverTimer(_In_ ULONGLONG dwTickCount)
{
  m_lastMouseLoc.x = INT_MIN;
  m_lastMouseLoc.y = INT_MIN;
  m_nLastMovementTime = dwTickCount;
  m_nLastScrollTime = dwTickCount;
}

#pragma warning(suppress: 26440)
DROPEFFECT CShellListCtrl2::PerformRightClickDragDrop(_In_ int nItem, _In_ DWORD dwDropDefault, _In_ DROPEFFECT dwDropList, _In_ HDROP hDrop, _In_ const CString& sToPath)
{
  CMenu popupMenu;
  if (!popupMenu.LoadMenu(IDR_SHELLCTRLS2_DROP))
  {
    ASSERT(FALSE);
    return DROPEFFECT_NONE;
  }
  CMenu* pPopup{popupMenu.GetSubMenu(0)};
  if (pPopup == nullptr)
  {
    ASSERT(FALSE);
    return DROPEFFECT_NONE;
  }
  switch (dwDropDefault)
  {
    case DROPEFFECT_COPY:
    {
      pPopup->SetDefaultItem(ID_SHELLCTRLS2_DROP_COPY);
      break;
    }
    case DROPEFFECT_MOVE:
    {
      pPopup->SetDefaultItem(ID_SHELLCTRLS2_DROP_MOVE);
      break;
    }
    default:
    {
      break;
    }
  }
  if ((dwDropList & DROPEFFECT_COPY) == 0)
    pPopup->EnableMenuItem(ID_SHELLCTRLS2_DROP_COPY, MF_BYCOMMAND | MF_DISABLED | MF_GRAYED);
  if ((dwDropList & DROPEFFECT_MOVE) == 0)
    pPopup->EnableMenuItem(ID_SHELLCTRLS2_DROP_MOVE, MF_BYCOMMAND | MF_DISABLED | MF_GRAYED);
  CPoint point;
  GetCursorPos(&point);
  DROPEFFECT deResult{DROPEFFECT_NONE};
  const int nID{pPopup->TrackPopupMenu(TPM_LEFTALIGN | TPM_LEFTBUTTON | TPM_RIGHTBUTTON | TPM_NONOTIFY | TPM_RETURNCMD, point.x, point.y, AfxGetMainWnd())};
  switch (nID)
  {
    case ID_SHELLCTRLS2_DROP_COPY:
    {
      if (SUCCEEDED(PerformDragDrop(nItem, DROPEFFECT_COPY, hDrop, sToPath)))
        deResult = DROPEFFECT_COPY;
      break;
    }
    case ID_SHELLCTRLS2_DROP_MOVE:
    {
      if (SUCCEEDED(PerformDragDrop(nItem, DROPEFFECT_MOVE, hDrop, sToPath)))
        deResult = DROPEFFECT_MOVE;
      break;
    }
    default:
    {
      break;
    }
  }
  return deResult;
}

#pragma warning(suppress: 26440)
HRESULT CShellListCtrl2::PerformDragDrop(_In_ int nItem, _In_ DWORD dwDropEffect, _In_ HDROP hDrop, _In_ const CString& sToPath)
{
  //Remove drop target highlighting before we perform the drop
  SelectDropTarget(-1);

  //Get the count of items dropped
  const UINT nCount{DragQueryFile(hDrop, UINT_MAX, nullptr, 0)};
  if (nCount == 0)
    return TYPE_E_OUTOFBOUNDS;

  //Create the IFileOperation object
  ATL::CComPtr<IFileOperation> pFileOperation;
  HRESULT hr{pFileOperation.CoCreateInstance(__uuidof(FileOperation), nullptr, CLSCTX_ALL)};
  if (FAILED(hr))
    return hr;

  //Setup the various settings for IFileOperation
  const CWnd* pParentWnd{AfxGetMainWnd()};
  if (pParentWnd != nullptr)
  {
    hr = pFileOperation->SetOwnerWindow(pParentWnd->GetSafeHwnd());
    if (FAILED(hr))
      return hr;
  }
  hr = pFileOperation->SetOperationFlags(FOF_NOCONFIRMMKDIR | FOF_ALLOWUNDO | FOFX_NOMINIMIZEBOX);
  if (FAILED(hr))
    return hr;

  //Queue up the Copy, Move or Delete operations
  m_bDropContainedDirectories = false;
  m_bDropContainedFiles = false;
  CComPtr<IShellItem> pDestItem;
#ifdef _UNICODE
#pragma warning(suppress: 26490)
  hr = SHCreateItemFromParsingName(sToPath, nullptr, IID_IShellItem, reinterpret_cast<void**>(&pDestItem));
#else
#pragma warning(suppress: 26490)
  hr = SHCreateItemFromParsingName(CStringW(sToPath), nullptr, IID_IShellItem, reinterpret_cast<void**>(&pDestItem));
#endif //#ifdef _UNICODE
  if (FAILED(hr))
    return hr;
  for (UINT i{0}; i<nCount; i++)
  {
    const UINT nCharSize{DragQueryFile(hDrop, i, nullptr, 0)};
    if (nCharSize > 0)
    {
      CString sSourceFile;
      DragQueryFile(hDrop, i, sSourceFile.GetBuffer(nCharSize), nCharSize + 1);
      sSourceFile.ReleaseBuffer();
      const DWORD dwAttributes{GetFileAttributes(sSourceFile)};
      if (dwAttributes != INVALID_FILE_ATTRIBUTES)
      {
        if (!m_bDropContainedFiles)
          m_bDropContainedFiles = ((dwAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0);
        if (!m_bDropContainedDirectories)
          m_bDropContainedDirectories = (dwAttributes & FILE_ATTRIBUTE_DIRECTORY);
      }
      CComPtr<IShellItem> pSourceItem;
    #ifdef _UNICODE
#pragma warning(suppress: 26490)
      hr = SHCreateItemFromParsingName(sSourceFile, nullptr, IID_IShellItem, reinterpret_cast<void**>(&pSourceItem));
    #else
#pragma warning(suppress: 26490)
      hr = SHCreateItemFromParsingName(CStringW(sSourceFile), nullptr, IID_IShellItem, reinterpret_cast<void**>(&pSourceItem));
    #endif //#ifdef _UNICODE
      if (FAILED(hr))
        return hr;
      if (dwDropEffect == DROPEFFECT_COPY)
        hr = pFileOperation->CopyItem(pSourceItem, pDestItem, nullptr, nullptr);
      else
        hr = pFileOperation->MoveItem(pSourceItem, pDestItem, nullptr, nullptr);
      if (FAILED(hr))
        return hr;
    }
  }

  //Finally do the file operations
  hr = pFileOperation->PerformOperations();
  if (FAILED(hr))
    return hr;

  //Notify the parent of the successful drop
  const CWnd* pParent{GetParent()};
  if (pParent != nullptr)
  {
#pragma warning(suppress: 26472)
    pParent->SendMessage(WM_SHELLLISTCTRL2_AFTER_OLEDROP, 0, static_cast<LPARAM>(nItem));
  }

  return S_OK;
}

void CShellListCtrl2::SelectDropTarget(_In_ int nItem)
{
  //Restore the selection from any previous drop target
  if (m_nDropTarget != -1)
  {
    if (m_bDropTargetWasSelected)
      SetItemState(m_nDropTarget, LVIS_SELECTED, LVIS_SELECTED);
    else
      SetItemState(m_nDropTarget, 0, LVIS_SELECTED);
  }

  //Remember the old selection state of the drop target before we select it
  if (nItem != -1)
  {
    m_bDropTargetWasSelected = (GetItemState(nItem, LVIS_SELECTED) != 0);
    //Select the drop target
    SetItemState(nItem, LVIS_SELECTED, LVIS_SELECTED);
  }

  //Remember the new drop target index
  m_nDropTarget = nItem;
}
