/*
Module : cpidl.h
Purpose: Simple C++ class called CPDIDL which provide a RAII class wrapper for LPITEMIDLIST aka PIDLs
Created: PJN / 25-07-2020
History: PJN / 25-07-2020 1. Initial creation
         PJN / 23-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable 
                          declarations.

Copyright (c) 2020 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __CPIDL_H__
#define __CPIDL_H__

#ifndef CPIDL_EXT_CLASS
#define CPIDL_EXT_CLASS
#endif //#ifndef CPIDL_EXT_CLASS


//////////////////// Includes /////////////////////////////////////////////////

#ifndef _VECTOR_
#pragma message("To avoid this message, put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

#ifndef __ATLCOM_H__
#pragma message("To avoid this message, please put atlcom.h in your pre compiled header (normally stdafx.h)")
#include <atlcom.h>
#endif //#ifndef __ATLCOM_H__

#ifndef _SHLOBJ_H_
#pragma message("To avoid this message, put shlobj.h in your pre compiled header (normally stdafx.h)")
#include <shlobj.h>
#endif //#ifndef _SHLOBJ_H_


//////////////////// Classes //////////////////////////////////////////////////

//Class which provides a RAII wrapper for a PIDL
class CPIDL_EXT_CLASS CPIDL
{
public:
//Constructors / Destructors
  CPIDL() noexcept : m_pPIDL{nullptr}
  {
  }

  CPIDL(const CPIDL& pidl) : CPIDL{}
  {
#pragma warning(suppress: 26447)
    const auto nSize{pidl.GetSize()};
    Create(nSize);
    CopyMemory(m_pPIDL, pidl.m_pPIDL, nSize);
  }

  CPIDL(CPIDL&& pidl) noexcept : CPIDL{}
  {
    m_pPIDL = pidl.m_pPIDL;
    pidl.m_pPIDL = nullptr;
  }

  ~CPIDL()
  {
#pragma warning(suppress: 26447)
    Free();
  }

//Methods
  CPIDL& operator=(const CPIDL& pidl)
  {
    if (this != &pidl)
    {
      const auto nSize{pidl.GetSize()};
      Create(nSize);
      CopyMemory(m_pPIDL, pidl.m_pPIDL, nSize);
    }
    return *this;
  }

#pragma warning(suppress: 26439)
  CPIDL& operator=(CPIDL&& pidl)
  {
    Free();
    m_pPIDL = pidl.m_pPIDL;
    pidl.m_pPIDL = nullptr;
    return *this;
  }

  _NODISCARD CPIDL operator+(const CPIDL& pidl) const
  {
    return Concat(*this, pidl);
  }

  void Attach(LPITEMIDLIST pPIDL) noexcept
  {
    Free();
    m_pPIDL = pPIDL;
  }

  _NODISCARD LPITEMIDLIST Detach() noexcept
  {
    LPITEMIDLIST pTempPIDL{m_pPIDL};
    m_pPIDL = nullptr;
    return pTempPIDL;
  }

  void Create(_In_ UINT cbSize) noexcept
  {
    Free();
    m_pPIDL = static_cast<LPITEMIDLIST>(CoTaskMemAlloc(cbSize));
    if (m_pPIDL != nullptr)
      ZeroMemory(m_pPIDL, cbSize);
  }

  void Free() noexcept
  {
    if (m_pPIDL != nullptr)
    {
      CoTaskMemFree(m_pPIDL);
      m_pPIDL = nullptr;
    }
  }

  _NODISCARD UINT GetItemCount() const
  {
    if (m_pPIDL == nullptr)
      return 0;
    UINT nCount{0};
    LPCSHITEMID pid{GetFirstItemID()};
    do
    {
      if (pid->cb)
        nCount++;
      GetNextItemID(pid);
    }
    while (pid->cb);
    return nCount;
  }

  _NODISCARD UINT GetSize() const
  {
    if (m_pPIDL == nullptr)
      return 0;
    UINT nSize{0};
    LPCSHITEMID pid{GetFirstItemID()};
    if (pid != nullptr)
    {
      do
      {
        nSize += pid->cb;
        GetNextItemID(pid);
      }
      while (pid->cb);
      nSize += sizeof(pid->cb); //Requires a 2 byte null terminator
    }
    return nSize;
  }

  void FromMemory(_In_reads_bytes_(nSize) const BYTE* pData, _In_ UINT nSize) noexcept
  {
    Free();
    if (nSize)
    {
      Create(nSize);
      CopyMemory(m_pPIDL, pData, nSize);
    }
  }

  _NODISCARD std::vector<BYTE> ToByteArray() const
  {
    std::vector<BYTE> data;
    const auto nSize{GetSize()};
    data.reserve(nSize);
#pragma warning(suppress: 26481 26490)
    data.insert(data.end(), reinterpret_cast<const BYTE*>(m_pPIDL), reinterpret_cast<const BYTE*>(m_pPIDL) + nSize);
    return data;
  }

  _NODISCARD LPSHITEMID GetFirstItemID() const noexcept
  {
#pragma warning(suppress: 26490)
    return reinterpret_cast<LPSHITEMID>(m_pPIDL);
  }

#pragma warning(suppress: 26429 26440)
  static void GetNextItemID(_Inout_ LPCSHITEMID& pid)
  {
    //Validate our parameters
    ASSERT(pid != nullptr);

#pragma warning(suppress: 26481 26490)
    reinterpret_cast<const BYTE*&>(pid) += pid->cb;
  }

#pragma warning(suppress: 26429 26440)
  static void GetNextItemID(_Inout_ LPSHITEMID& pid)
  {
    //Validate our parameters
    ASSERT(pid != nullptr);

#pragma warning(suppress: 26481 26490)
    reinterpret_cast<LPBYTE&>(pid) += pid->cb;
  }

  _NODISCARD LPCSHITEMID GetItemID(_In_ int nIndex) const
  {
    LPCSHITEMID pid{GetFirstItemID()};
    for (int i{0}; i<nIndex; i++)
      GetNextItemID(pid);
    return pid;
  }

  _NODISCARD LPSHITEMID GetItemID(_In_ int nIndex)
  {
    LPSHITEMID pid{GetFirstItemID()};
    ASSERT(pid != nullptr);
    for (int i{0}; i<nIndex; i++)
      GetNextItemID(pid);
    return pid;
  }

  HRESULT GetParent(_Inout_ CPIDL& parent) const
  {
    if (m_pPIDL == nullptr)
      return E_FAIL;
    const auto nCount{GetItemCount()};
    if (nCount == 0) //It does not really make sense to call this function for the desktop. Just fail the call.
      return E_FAIL;
    if (nCount == 1)
    {
      parent.Free();
      return SHGetKnownFolderIDList(FOLDERID_Desktop, 0, nullptr, parent);
    }

    USHORT nParentSize{0};
    LPSHITEMID pid{GetFirstItemID()};
    ASSERT(pid != nullptr);
    UINT i{0};
    do
    {
      nParentSize += pid->cb;
      GetNextItemID(pid);
      i++;
    }
    while (i < (nCount - 1));
    parent.Create(nParentSize + 2);
    CopyMemory(reinterpret_cast<LPBYTE>(parent.m_pPIDL), reinterpret_cast<const BYTE*>(m_pPIDL), nParentSize);
    return S_OK;
  }

  _NODISCARD bool IsDescendant(const CPIDL& other) const
  {
    if ((m_pPIDL == nullptr) || (other.m_pPIDL == nullptr))
      return false;
    bool bContinue{true};
    CPIDL parentThis;
    CPIDL currentThis{*this};
    do
    {
      const HRESULT hr{currentThis.GetParent(parentThis)};
      if (SUCCEEDED(hr))
      {
        if (other == parentThis)
          return true;
        currentThis = parentThis;
      }
      else
        bContinue = false;
    }
    while (bContinue);
    return false;
  }

  _NODISCARD bool IsAncestor(const CPIDL& other) const
  {
    if ((m_pPIDL == nullptr) || (other.m_pPIDL == nullptr))
      return false;
    bool bContinue{true};
    CPIDL parentOther;
    CPIDL currentOther{other};
    do
    {
      const HRESULT hr{currentOther.GetParent(parentOther)};
      if (SUCCEEDED(hr))
      {
        if (*this == parentOther)
          return true;
        currentOther = parentOther;
      }
      else
        bContinue = false;
    }
    while (bContinue);
    return false;
  }

  HRESULT FromPath(_In_z_ LPCTSTR pszPath)
  {
    ATL::CComPtr<IShellFolder> pDesktopFolder;
    HRESULT hr{SHGetDesktopFolder(&pDesktopFolder)};
    if (FAILED(hr))
      return hr;

    //IShellFolder::ParseDisplayName requires the file name be in Unicode
    CStringW sPath{pszPath};

    //Convert the path to an ITEMIDLIST
    Free();
    ULONG dwAttributes{0};
    hr = pDesktopFolder->ParseDisplayName(nullptr, nullptr, sPath.GetBuffer(), nullptr, &m_pPIDL, &dwAttributes);
    sPath.ReleaseBuffer();
    return hr;
  }

  bool ToPath(_Inout_ CString& sPath) const
  {
    const BOOL bRes{SHGetPathFromIDList(m_pPIDL, sPath.GetBuffer(_MAX_PATH))};
    sPath.ReleaseBuffer();
    return bRes;
  }

  void Split(_Inout_ CPIDL& parent, _Inout_ CPIDL& object) const
  {
    //Default the output parameters
    parent = CPIDL{};
    object = CPIDL{};

    if (m_pPIDL == nullptr) //Nothing to do for a null pidl
      return;

    //Find the last ITEMID in this item and also calculate 
    //the size of the parent PIDL at the same time
    UINT nSize{0};
    LPCSHITEMID pid{GetFirstItemID()};
    LPCSHITEMID pidLast{nullptr};
    do
    {
      pidLast = pid;
      nSize += pid->cb;
      GetNextItemID(pid);
    }
    while (pid->cb);
    nSize -= pidLast->cb;

    //Create the parent
    parent.Create(nSize + sizeof(pid->cb));
    CopyMemory(parent.m_pPIDL, m_pPIDL, nSize);

    //Create the object
    nSize = pidLast->cb + sizeof(pidLast->cb);
    object.Create(nSize);
    CopyMemory(object.m_pPIDL, pidLast, nSize);
  }

  _NODISCARD bool operator==(const CPIDL& pidl) const
  {
    if ((m_pPIDL == nullptr) && (pidl.m_pPIDL == nullptr)) //Both empty, return true
      return true;
    else if ((m_pPIDL == nullptr) && (pidl.m_pPIDL != nullptr)) //this empty, then return false
      return false;
    else if ((m_pPIDL != nullptr) && (pidl.m_pPIDL == nullptr)) //other pidl empty, then return false
      return false;
    const auto nSize{GetSize()};
    if (nSize != pidl.GetSize())
     return false;
    return memcmp(m_pPIDL, pidl.m_pPIDL, nSize) == 0;
  }

  _NODISCARD bool operator!=(const CPIDL& pidl) const
  {
    return !operator==(pidl);
  }

  operator LPITEMIDLIST&() noexcept
  {
    return m_pPIDL;
  }

#pragma warning(suppress: 26440)
  operator LPITEMIDLIST*()
  {
    ASSERT(m_pPIDL == nullptr); //If this assert files, this usually indicates a bug. If this really is what's needed
                                //then take the address of the m_pPIDL member variable explicitly.
    return &m_pPIDL;
  }

  operator LPCITEMIDLIST() const noexcept
  {
    return m_pPIDL;
  }

  operator LPCITEMIDLIST*() const noexcept
  {
#pragma warning(suppress: 26492)
    return const_cast<LPCITEMIDLIST*>(&m_pPIDL);
  }

  operator bool() const noexcept
  {
    return (m_pPIDL != nullptr);
  }

//Member variables
  LPITEMIDLIST m_pPIDL;

protected:
//Methods
  _NODISCARD static CPIDL Concat(const CPIDL& a, const CPIDL& b)
  {
    if (!a && !b) //Both empty, return empty
      return CPIDL{};
    else if (!a) //A empty, then return B
      return b;
    else if (!b) //B empty, then return A
      return a;

#pragma warning(suppress: 26472)
    const UINT nSize1{static_cast<UINT>(a.GetSize() - sizeof(a.m_pPIDL->mkid.cb))}; //Do not include the null terminator from the first pidl
    const UINT nSize2{b.GetSize()};
    CPIDL result;
#pragma warning(suppress: 26486)
    result.Create(nSize1 + nSize2);
#pragma warning(suppress: 26486)
    CopyMemory(result.m_pPIDL, a.m_pPIDL, nSize1);
#pragma warning(suppress: 26481)
    CopyMemory(reinterpret_cast<BYTE*>(result.m_pPIDL) + nSize1, b.m_pPIDL, nSize2);
    return result;
  }
};


#endif //#ifndef __CPIDL_H__
